/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef HSSR_PAR_COMPRESS_HPP
#define HSSR_PAR_COMPRESS_HPP

template<typename T, typename S> void HSS_tree<T,S>::HSSR_par_Compress(T *A, int *descA, S tol, int dmin, int dmax, int dlim, int dinc, int *deff, int *r, gridinfo *ginfo, T *Rr, T *Rc, T *Sr, T *Sc, int *iodescRS, void (*elements)(void *, int *, int *, T *, int *), bool dosync) {
  /* Parallel HSS Compression of a dense matrix with randomized sampling.
   * A is a dense n x n matrix.
   * dmin is the number of random vectors to be used; it is used
   * at the beginning of the compression. If, during compression,
   * the HSS rank is found to be too close to the number of
   * random vectors, the number of random vectors is increased.
   * dlim is how close the rank is allowed to be from the number
   * of random vectors.
   * dmax is the maximum number of random vectors to be used.
   * dinc is the increment (number of additional random vectors
   * when expanded).
   * deff is the effective number of random vectors used.
   * r is the maximum rank found during the compression (output).
   * Rr, Rc, Sr, Sr are random vectors and samplings of A (see below);
   * they are either provided (non-NULL) or computed by the routine.
   * The matrix A can be provided as:
   *  - An explicit matrix,
   *  - An element access routine and explicit samples Sr and Sc,
   *  - An element access routine and a matrix-vector product.
   *
   * Algorithm:
   * Step1: Rr and Rc are two n x d random matrices.
   * Step2: Sr and Sc are two n x d samples (resp. row and column) of A.
   * Step3: postorder traversal of the HSS tree; at each node:
   *   a/ Extract a local sample.
   *   b/ Compress with interpolative decomposition.
   *   c/ Update random vectors and indices.
   *
   * A basic proportional mapping is applied to the HSS tree. Each process owns
   * a serial subtree (which might be empty) and a chain of nodes that are
   * shared with other processes. We call "parallel leaves":
   *  - The root nodes of the serial subtrees (nodes mapped on one process
   *  with a parent mapped on two processes or more).
   *  - The leaves that happen to be mapped on more than one process (nodes
   *  mapped on two processes or more without a node below them).
   * In other words, "parallel leaves" are the leaves of the parallel part of
   * the tree. The set of parallel leaves induces a clustering of [1 n].
   * In terms of implementation, we know that a node is a parallel leaf when
   *      procs[1]==procs[0] and nprow>0 (nodes inside serial subtrees have nprow=-1)
   *   or procs[1]> procs[0] and nr_children=0.
   * It is important to distinguish between two different things:
   *   - the set of processes an node is mapped on.
   *   - the set of processes that *work* at a node.
   * The difference comes from the fact that we use a BLACS grid at each node
   * and we try to maintain squarish grids. For example, if a node is mapped on
   * 37 processes, then only 36 actually belong to the (6x6) grid and own
   * pieces of the data associated with the node. The 37th process is idle,
   * but it may be active at some descendants or ancestors of the node. The
   * code is as general as it can be.
   *
   * Data distribution:
   *   - A is distributed in a 2D block-cyclic format.
   *   - Sr, Sc, Rr and Rc are first created with 2D-block format and
   *     are then redistributed to conform with the clustering of [1 n]
   *     defined by the parallel leaves of the HSS tree.
   *   - The diagonal blocks of A (corresponding to parallel leaves) are
   *     distributed so that each process has all the information needed to
   *     traverse its serial subtree without communicating with the other
   *     processes.
   *   - The off-diagonal blocks of A (that are later compressed into B12, B21)
   *     are also distributed according to the mapping of the tree, so
   *     that when a parallel node is processed, the processes that handle
   *     this node have all the information they need before they start.
   *   - A is duplicated and the copy is distributed using the two rules
   *     above; A is then left untouched.
   *
   * Example: consider a 6x6 matrix. The HSS tree has 4 leaves that correspond
   * to the clustering {[1 2],[3],[4 5],[6]} of [1 6]. The tree is mapped on
   * 4 processes as follows (it is not proportional mapping but it covers all
   * the situations that can arise):
   *
   *                    [1 6]:P0+P1+P2+P3
   *                  /       \
   *                /           \
   *              /               \
   *      P0:[1 3]                 [4 6]:P1+P2+P3
   *         /   \                 /   \
   *  P0:[1 2]   3:P0    P1+P2:[4 5]   6: P3
   *
   * P0 has a serial subtree of size 3 (the 3 leftmost nodes). P1 and P2 have
   * an empty serial subtree. P3 has a serial subtree of size 1. There are
   * three parallel leaves: [1 3], [4 5] and [6]. The first one is not a leaf
   * of the HSS tree, but it's the root node of a serial subtree; the second
   * one is a leaf of the HSS tree that happens to be mapped on more than 1
   * process.
   *
   * Each process is given the following pieces of A:
   *   - P0 has A(1:3,1:3) that it needs to process its subtree, because it
   *     contains A(1:2,1:2) and A(3,3), that are needed to process the first 2
   *     nodes ("D" generators); P0 also has A(1:2,3) and A(3,1:2) that are
   *     supersets of what is needed for the third node (B12 and B21
   *     generators).
   *   - P1 and P2 share A(4:5,4:5) with a 2D block cyclic distribution. This is
   *     what is needed to process the corresponding leaf. They also share
   *     A(4:5,6) and A(6,4:5) with P3. This is what is needed to process
   *     node [4 6] (supersets of B12 and B21).
   *   - P3 has A(6,6) and shares A(4:5,6) and A(6,4:5) with P1 and P2.
   *
   * Each process is given the following pieces of X=Rr/Rc/Sr/Sc:
   *   - P0 has X(1:3,:), that it needs to process it serial subtree.
   *   - P1 and P2 share X(4:5,:) and they share X(6,:) with P3.
   *   - P3 has X(6,:).
   *
   * Limitations:
   *   - The algorithm assumes that the HSS tree is binary.
   *   - All the MPI ranks are assumed to call the routine.
   */

  int n;
  int d;
  int mb, nb;
  int locrRS, loccRS, locr, locc;
  int idummy, idummy1, idummy2, ierr;
  int *desc;
  int *descRS, *descRSr=NULL, *descRSc=NULL;
  int *descRS_upd, *descRSr_upd, *descRSc_upd;
  int *descD, **descA12, **descA21;
  int *tmpI, *tmpJ;
  int nblevels, mymaxlevel;
  int *nb_calls_level;
  T *Rr_upd, *Rc_upd, *Sr_upd, *Sc_upd;
  T *D, *tmp;
  T **A12, **A21;
  int i, loc_i, j, k;
  int istart, iend;
  int ib, ie, b, br, bc;
  int radd;
  int myid, np;
  int proc_b, proc_g, proc_e;
  bool was_failed;
  bool RSprovided;
  MPI_Request request;
  MPI_Status status;
  gridinfo tmpinfo;
  HSS_tree<T,S> loc_htree;
  HSS_node<T,S> *child1, *child2;
  double t, tRand, tGEMM;

#if defined(debug)
  int idebug, mdebug, ndebug, rdebug, cdebug;
  HSS_node<T,S> *nodedebug;
#endif

  /* Some initializations */
  n=a_dim;
  mb=ginfo->mb;
  nb=ginfo->nb;
  *r=0;
  MPI_Comm_rank(comm,&myid);
  MPI_Comm_size(comm,&np);
  time_Rand=0;
  time_GEMM=0;
  time_ID=0;

  /* Synchronous traversal: if the element access routine provided by the user
   * (if any) is global, we need to make sure that, at each level, calls to
   * the routine are done one after another, and that every process participate
   * to every call. Therefore, at a given node (level), a process has to:
   *  1/ perform a bunch of calls to "elements" that are useful to its previous
   *    siblings.
   *  2/ perform a call to "elements" to get D (if leaf) or two calls to get B12
   *    and B21 (if interior node).
   *  3/ perform a bunch of calls to "elements" that are useful to its following
   *    siblings.
   *  Also, when the tree is imbalanced (not complete), there are levels for which
   *  a given process does not own any node. In that case, that process also has
   *  to perform calls to elements.
   *
   *  nb_calls_level[l] is the total number of calls to "elements" at level l.
   *  nodes[i]->nb_calls_prev is the number of calls to "elements" that have to
   *  be performed at step 1/ for node i.
   *  nodes[i]->nb_calls_next is the number of calls to "elements" that have to
   *  be performed at step 2/ for node i.
   *
   *  Note that it could happen that "elements" does not have any global synchronization.
   *  This potentially messes up the order of messages and we enforce a barrier after
   *  every call to prevent this.
   */
  if(dosync) {
    nblevels=0;
    nodes[dim_tree-1]->level=0;
    for(i=dim_tree-1;i>=0;i--)
      if(nodes[i]->nr_children!=0) {
        nodes[i]->children[0]->level=nodes[i]->level+1;
        nodes[i]->children[1]->level=nodes[i]->level+1;
        nblevels=std::max(nblevels,nodes[i]->children[0]->level);
      }
    nblevels++; /* Levels are 0-based */

    nb_calls_level=new int[nblevels]();
    for(i=0;i<dim_tree;i++) {
      nodes[i]->nb_calls_prev=nb_calls_level[nodes[i]->level];
      if(nodes[i]->nr_children==0)
        nb_calls_level[nodes[i]->level]+=1;
      else
        nb_calls_level[nodes[i]->level]+=2;
    }
    for(i=0;i<dim_tree;i++)
      if(nodes[i]->nr_children==0)
        nodes[i]->nb_calls_next=nb_calls_level[nodes[i]->level]-nodes[i]->nb_calls_prev-1;
      else
        nodes[i]->nb_calls_next=nb_calls_level[nodes[i]->level]-nodes[i]->nb_calls_prev-2;

#if defined(debug)
    if(!myid) {
      for(i=0;i<dim_tree;i++)
        std::cout << "Node " << i << " at level " << nodes[i]->level << "; prev=" << nodes[i]->nb_calls_prev << " next=" << nodes[i]->nb_calls_next << std::endl;
      std::cout << "nblevels=" << nblevels << std::endl;
    }
#endif
  }

  //TODO !!! For adaptive, need to modify nb_calls_* according to states...

  /* Step 1: generating random vectors Rr and Rc.
   * +
   * Step 2: generating samples: Sr=A*Rr, Sc=A'*Rc. */
  d=dmin;
  StrumpackRand rgen(time(NULL)+myid);
  if(Rr==NULL && Rc==NULL && Sr==NULL && Sc==NULL) {
    RSprovided=false;
    generateSamples(&Rr,&Sr,&Rc,&Sc,&descRS,n,d,A,descA,ginfo,&rgen,comm,&tRand,&tGEMM);
    time_Rand+=tRand;
    time_GEMM+=tGEMM;
  } else {
    RSprovided=true;
    descRS=new int[BLACSCTXTSIZE];
    copyDesc(iodescRS,descRS);
  }

  /* Step 3: parallel postorder traversal of the tree */

  /* The first step is to redistribute A, Rr, Rc, Sr, Sc according
   * to the tree structure.
   * par_lvl is the number of parallel *interior* nodes in which
   * myid is involved.
   * Rr, Rc, Sr and Sc are redistributed so that a process only owns
   * the piece of Rr/Rc/Sr/Sc that corresponds to the (unique) parallel
   * leaf on which it works.
   * D is the diagonal block of A that corresponds to the parallel leaf
   * that myid owns.
   * A12 and A21 (par_lvl pointers to dense blocks) are a distribution
   * of the off-diagonal blocks of A that conforms to the HSS tree.
   * If node i is the j-th parallel interior node that myid has to process, and
   * node i has two children c1 and c2 that correspond to the clusters [ib1
   * ie1] and [ib2 ie2], then A12[j] (resp. A21[j]) is a local piece of
   * A(ib1:ie1,ib2:ie2) (resp. A(ib2:ie2,ib1:ie1)).
   * We also return the descriptors for D and all the blocks in A12 and A21.
   */

  if(elements==NULL) {
  /* Call the redistribution routine for A, Rr, Rc, Sr, Sc */
    A12 = new T*[par_lvl];
    descA12 = new int*[par_lvl];
    A21 = new T*[par_lvl];
    descA21 = new int*[par_lvl];

    t=MPI_Wtime();
    distMatrixTree(A,descA,par_lvl,&D,&descD,A12,descA12,A21,descA21,ginfo,nodes[dim_tree-1]->allctxt);
    time_Dist=MPI_Wtime()-t;
  } else {
    D=NULL;
  }

  distVectorTree(Rr,descRS,d,&tmp,&desc,ginfo,nodes[dim_tree-1]->allctxt);
  if(!RSprovided)
    delete[] Rr;
  Rr=tmp;

  delete []desc;
  distVectorTree(Rc,descRS,d,&tmp,&desc,ginfo,nodes[dim_tree-1]->allctxt);
  if(!RSprovided)
    delete[] Rc;
  Rc=tmp;

  delete []desc;
  distVectorTree(Sr,descRS,d,&tmp,&desc,ginfo,nodes[dim_tree-1]->allctxt);
  if(!RSprovided)
    delete[] Sr;
  Sr=tmp;

  delete []desc;
  distVectorTree(Sc,descRS,d,&tmp,&desc,ginfo,nodes[dim_tree-1]->allctxt);
  if(!RSprovided)
    delete[] Sc;
  Sc=tmp;
  delete[] descRS;
  descRS=desc;

#if defined(debug)
  if(elements==NULL) {
    for(idebug=0;idebug<dim_tree;idebug++) {
      nodedebug=nodes[idebug];
      if(   nodedebug->isParLeaf
         && nodedebug->procs[0]<=myid
         && nodedebug->procs[1]>=myid) {
        ndebug=nodedebug->tree_ie-nodedebug->tree_ib+1;
        rdebug=numroc_(&ndebug,&mb,&nodedebug->myrow,&IZERO,&nodedebug->nprow);
        cdebug=numroc_(&ndebug,&nb,&nodedebug->mycol,&IZERO,&nodedebug->npcol);
        printMatrix("D: "+SSTR(myid)+","+SSTR(idebug),D,rdebug,cdebug,rdebug);
      }
    }

    for(idebug=0;idebug<par_lvl;idebug++) {
      nodedebug=nodes[par_nodes[idebug]];
      mdebug=nodedebug->children[0]->tree_ie-nodedebug->children[0]->tree_ib+1;
      ndebug=nodedebug->children[1]->tree_ie-nodedebug->children[1]->tree_ib+1;
      rdebug=numroc_(&mdebug,&mb,&nodedebug->myrow,&IZERO,&nodedebug->nprow);
      cdebug=numroc_(&ndebug,&nb,&nodedebug->mycol,&IZERO,&nodedebug->npcol);
      printMatrix("A12: "+SSTR(myid)+","+SSTR(idebug),A12[idebug],rdebug,cdebug,rdebug);
      rdebug=numroc_(&ndebug,&mb,&nodedebug->myrow,&IZERO,&nodedebug->nprow);
      cdebug=numroc_(&mdebug,&nb,&nodedebug->mycol,&IZERO,&nodedebug->npcol);
      printMatrix("A21: "+SSTR(myid)+","+SSTR(idebug),A21[idebug],rdebug,cdebug,rdebug);
    }
  }

  for(idebug=0;idebug<dim_tree;idebug++) {
    nodedebug=nodes[idebug];
    if(   nodedebug->isParLeaf
       && nodedebug->procs[0]<=myid
       && nodedebug->procs[1]>=myid) {
      ndebug=nodedebug->tree_ie-nodedebug->tree_ib+1;
      rdebug=numroc_(&ndebug,&mb,&nodedebug->myrow,&IZERO,&nodedebug->nprow);
      cdebug=numroc_(&d,&nb,&nodedebug->mycol,&IZERO,&nodedebug->npcol);
      printMatrix("Rr: "+SSTR(myid)+","+SSTR(idebug),Rr,rdebug,cdebug,rdebug);
    }
  }

  for(idebug=0;idebug<dim_tree;idebug++) {
    nodedebug=nodes[idebug];
    if(   nodedebug->isParLeaf
       && nodedebug->procs[0]<=myid
       && nodedebug->procs[1]>=myid) {
      ndebug=nodedebug->tree_ie-nodedebug->tree_ib+1;
      rdebug=numroc_(&ndebug,&mb,&nodedebug->myrow,&IZERO,&nodedebug->nprow);
      cdebug=numroc_(&d,&nb,&nodedebug->mycol,&IZERO,&nodedebug->npcol);
      printMatrix("Sr: "+SSTR(myid)+","+SSTR(idebug),Sr,rdebug,cdebug,rdebug);
    }
  }

  for(idebug=0;idebug<dim_tree;idebug++) {
    nodedebug=nodes[idebug];
    if(   nodedebug->isParLeaf
       && nodedebug->procs[0]<=myid
       && nodedebug->procs[1]>=myid) {
      ndebug=nodedebug->tree_ie-nodedebug->tree_ib+1;
      rdebug=numroc_(&ndebug,&mb,&nodedebug->myrow,&IZERO,&nodedebug->nprow);
      cdebug=numroc_(&d,&nb,&nodedebug->mycol,&IZERO,&nodedebug->npcol);
      printMatrix("Rc: "+SSTR(myid)+","+SSTR(idebug),Rc,rdebug,cdebug,rdebug);
    }
  }

  for(idebug=0;idebug<dim_tree;idebug++) {
    nodedebug=nodes[idebug];
    if(   nodedebug->isParLeaf
       && nodedebug->procs[0]<=myid
       && nodedebug->procs[1]>=myid) {
      ndebug=nodedebug->tree_ie-nodedebug->tree_ib+1;
      rdebug=numroc_(&ndebug,&mb,&nodedebug->myrow,&IZERO,&nodedebug->nprow);
      cdebug=numroc_(&d,&nb,&nodedebug->mycol,&IZERO,&nodedebug->npcol);
      printMatrix("Sc: "+SSTR(myid)+","+SSTR(idebug),Sc,rdebug,cdebug,rdebug);
    }
  }
#endif

  /* Start with a clean tree */
  for(i=0;i<dim_tree;i++)
    nodes[i]->state=UNTOUCHED;
  d=dmin-dinc;
restart:
  /* If this is the first time, d=dmin. Otherwise, d=dmin+x*dinc */
  d=d+dinc;

  /* Synchronized traversal: handle levels for which a given process
   * does not own any node.
   */
  if(dosync) {
    mymaxlevel=0;
    for(i=0;i<dim_tree;i++) {
      proc_b=nodes[i]->procs[0];
      proc_e=nodes[i]->procs[1];
      if(proc_e>=myid && proc_b<=myid)
        mymaxlevel=std::max(mymaxlevel,nodes[i]->level);
    }
    for(i=nblevels-1;i>mymaxlevel;i--) {
      participate_to_calls(elements,nb_calls_level[i],comm);
    }
  }


  /* Second, each process traverses its serial subtree, without
   * taking care of what the other processes do.
   *
   * A local HSS tree is created by taking only nodes subtree[2*myid]...
   * subtree[2*myid+1] of htree; indices associated with each node are shifted
   * to that they start at 1 and the sequential compression routine can be
   * called.
   *
   * If there are more than one process, the sequential routine is called for a
   * subtree of the complete tree. At the root node of this subtree, RRQR and
   * other operations must be performed, which is not the case when the
   * sequential subtree is equal to whole tree. By calling with doRoot=np-1, we
   * ensure that the root node of the serial subtree is correctly processed
   * (nothing to do if np==1, compression and updates if np>1).
   *
   * After the sequential compression, the indices are shifted back to conform
   * with the clustering of the global tree, and the information in the local
   * tree is passed to the global tree. (not all the fields in the local tree
   * are valid for the global tree, e.g. the mapping, so we cannot simply point
   * to the nodes).
   *
   * The sequential routine does not actually compress Rr/Rc/Sr/Sc (they still
   * have n rows at the end); they are compressed here (only the first
   * subtree[iend]->u/v_rank are kept).
   *
   * We also need to initialize descriptors for Rr/Sc and Rc/Sr that now have
   * different sizes (resp v_rank x d and u_rank x d). descRSc is for Rr and Sc,
   * and descRSr is for Rc and Sr.
   */
  if(subtree[2*myid]>=0) {
    istart=subtree[2*myid];
    iend=subtree[2*myid+1];

    if(d==dmin) {
      /* First time */
      extractLocalSubtree(&loc_htree);
      loc_htree.comm=comm;
      loc_htree.obj=obj;
      loc_htree.nodes[loc_htree.dim_tree-1]->context=nodes[iend]->context;
    }
    if(elements==NULL)
      if(loc_htree.a_dim!=descD[2] || loc_htree.a_dim!=descRS[2]) {
        std::cout << "Internal error in HSSR_par_Compress " << loc_htree.a_dim << " " << descD[1] << std::endl;
        exit(1);
      }

    /* Run the serial compression on the subtree */
    for(i=istart,loc_i=0;i<=iend;i++,loc_i++)
      loc_htree.nodes[loc_i]->state=nodes[i]->state;
    if(dosync) {
      for(i=istart,loc_i=0;i<=iend;i++,loc_i++) {
        loc_htree.nodes[loc_i]->nb_calls_prev=nodes[i]->nb_calls_prev;
        loc_htree.nodes[loc_i]->nb_calls_next=nodes[i]->nb_calls_next;
      }
    }
    if(d==dmin) {
      /* First traversal of the subtree */
      loc_htree.HSSR_seq_Compress(D,tol,dmin,dmax,dlim,dinc,deff,&radd,np-1,Rr,Rc,Sr,Sc,&rgen,elements,nodes[istart]->tree_ib-1,dosync);
      *r=std::max(*r,radd);
    } else {
      if(loc_htree.nodes[loc_htree.dim_tree-1]->state==COMPRESSED) {
        /* Subtree was entirely compressed; we only need to apply updates
         * to S_upd, R_upd. We pass dmin=dinc because R/S_upd have dinc columns,
         * not the total number of random vectors. Internally, the routine
         * is going to update columns d-dinc...d-dinc+dinc of the samples,
         * with d=dmin=dinc, i.e., columns 0...dinc=all the columns of the
         * _upd array.
         */
        was_failed=false;
        loc_htree.HSSR_seq_Compress(D,tol,dinc,dmax,dlim,dinc,deff,&radd,np-1,Rr_upd,Rc_upd,Sr_upd,Sc_upd,&rgen,elements,nodes[istart]->tree_ib-1,dosync);
     } else {
        /* There was a failure in the subtree; merge updated columns. */
        was_failed=true;
        tmpinfo.ctxt=nodes[iend]->context;
        tmpinfo.nprow=nodes[iend]->nprow;
        tmpinfo.npcol=nodes[iend]->npcol;
        tmpinfo.myrow=nodes[iend]->myrow;
        tmpinfo.mycol=nodes[iend]->mycol;
        tmpinfo.mb=ginfo->mb;
        tmpinfo.nb=ginfo->nb;
        updateColumns(&Rr,loc_htree.a_dim,d-dinc,&descRS,Rr_upd,dinc,descRS_upd,&tmpinfo);
        delete[] Rr_upd;
        updateColumns(&Rc,loc_htree.a_dim,d-dinc,&descRS,Rc_upd,dinc,descRS_upd,&tmpinfo);
        delete[] Rc_upd;
        updateColumns(&Sr,loc_htree.a_dim,d-dinc,&descRS,Sr_upd,dinc,descRS_upd,&tmpinfo);
        delete[] Sr_upd;
        updateColumns(&Sc,loc_htree.a_dim,d-dinc,&descRS,Sc_upd,dinc,descRS_upd,&tmpinfo);
        delete[] Sc_upd;
        loc_htree.HSSR_seq_Compress(D,tol,d,dmax,dlim,dinc,deff,&radd,np-1,Rr,Rc,Sr,Sc,&rgen,elements,nodes[istart]->tree_ib-1,dosync);
        *r=std::max(*r,radd);
      }
    }

#if defined(debug)
     int debugn=loc_htree.a_dim;
     T *debugX=new T[debugn*debugn];
     T *debugR=new T[debugn*debugn];
     for(int debugj=0;debugj<debugn;debugj++)
        for(int debugi=0;debugi<debugn;debugi++) {
          debugX[debugj*debugn+debugi]=(debugi==debugj)?ONE:ZERO;
          debugR[debugj*debugn+debugi]=D[debugj*debugn+debugi];
        }
      loc_htree.HSS_seq_Product('N',ONE,debugX,debugn,NONE,debugR,0);
      S *debugwork=new S[debugn];
      std::cout << "Serial HSS compression: ||HSS(A)*I-A||_fro/||A||_fro=" << lange('F',debugn,debugn,debugR,debugn,debugwork) << std::endl;
      delete[] debugwork;
      delete[] debugX;
      delete[] debugR;
#endif

    /* Copy/point info from serial subtree to global tree */
    for(i=istart,loc_i=0;i<=iend;i++,loc_i++) {
      if(loc_htree.nodes[loc_i]->state==COMPRESSED && i!=dim_tree-1) {
        nodes[i]->U_E=loc_htree.nodes[loc_i]->U_E;
        nodes[i]->U_perm=loc_htree.nodes[loc_i]->U_perm;
        nodes[i]->u_rank=loc_htree.nodes[loc_i]->u_rank;
        nodes[i]->u_rows=loc_htree.nodes[loc_i]->u_rows;
        nodes[i]->Jr=loc_htree.nodes[loc_i]->Jr;
        nodes[i]->Ir=new int[loc_htree.nodes[loc_i]->u_rank];
        for(k=0;k<loc_htree.nodes[loc_i]->u_rank;k++)
          nodes[i]->Ir[k]=loc_htree.nodes[loc_i]->Ir[k]+nodes[istart]->tree_ib-1;
        nodes[i]->V_E=loc_htree.nodes[loc_i]->V_E;
        nodes[i]->V_perm=loc_htree.nodes[loc_i]->V_perm;
        nodes[i]->v_rank=loc_htree.nodes[loc_i]->v_rank;
        nodes[i]->v_rows=loc_htree.nodes[loc_i]->v_rows;
        nodes[i]->Jc=loc_htree.nodes[loc_i]->Jc;
        nodes[i]->Ic=new int[loc_htree.nodes[loc_i]->v_rank];
        for(k=0;k<loc_htree.nodes[loc_i]->v_rank;k++)
          nodes[i]->Ic[k]=loc_htree.nodes[loc_i]->Ic[k]+nodes[istart]->tree_ib-1;
        nodes[i]->numrand=loc_htree.nodes[loc_i]->numrand;
      }
      if(loc_htree.nodes[loc_i]->state==COMPRESSED || loc_htree.nodes[loc_i]->state==PARTIALLY_COMPRESSED) {
        nodes[i]->B12=loc_htree.nodes[loc_i]->B12;
        nodes[i]->B21=loc_htree.nodes[loc_i]->B21;
        nodes[i]->D=loc_htree.nodes[loc_i]->D;
      }
      nodes[i]->state=loc_htree.nodes[loc_i]->state;
    }
    time_Rand+=loc_htree.time_Rand;
    time_GEMM+=loc_htree.time_GEMM;
    time_ID+=loc_htree.time_ID;

    if(np>1) {
      /* Compression of the random vectors and samples.
       * This is done only in parallel to go to the upper
       * levels. Otherwise (1 process), the whole compression
       * is already done at this point and we only need to
       * free the random vectors and samples (which is done at
       * the end of this routine).
       *
       * This is done only if the serial subtree is successfully compressed
       * (root node has state COMPRESSED), otherwise there is no _upd vectors
       * (this is either the first traversal, or a restart because the serial
       * subtree failed, in which case update vectors have been merged into
       * R/S. If d==dmin, this is the first time compression was called; R/S
       * are updated. Otherwise, if was_failed is true, this is a restart
       * because the sequential subrtee, so now that the tree is successfully
       * compressed, S/R are compressed. If was_failed isfalse, this is a
       * restart but not because of the sequential subtree, so there are some
       * _upd vectors to be compressed.
       */

      if(nodes[iend]->state==COMPRESSED) {
        if(d==dmin || was_failed) {
          /* Compress Rr and Sc, and initialize descRSc */
          b=nodes[iend]->v_rank;
          tmp=new T[b*d];
          lacpy('N',b,d,Rr,loc_htree.a_dim,tmp,b);
          delete[] Rr;
          Rr=tmp;

          tmp=new T[b*d];
          lacpy('N',b,d,Sc,loc_htree.a_dim,tmp,b);
          delete[] Sc;
          Sc=tmp;

          descRSc=new int[BLACSCTXTSIZE];
          idummy=std::max(1,numroc_(&b,&mb,&nodes[iend]->myrow,&IZERO,&nodes[iend]->nprow));
          descinit_(descRSc,&b,&d,&mb,&nb,&IZERO,&IZERO,&(nodes[iend]->context),&idummy,&ierr);

          /* Compress Rc and Sr, and initialize descRSr */
          b=nodes[iend]->u_rank;
          tmp=new T[b*d];
          lacpy('N',b,d,Rc,loc_htree.a_dim,tmp,b);
          delete[] Rc;
          Rc=tmp;

          tmp=new T[b*d];
          lacpy('N',b,d,Sr,loc_htree.a_dim,tmp,b);
          delete[] Sr;
          Sr=tmp;

          descRSr=new int[BLACSCTXTSIZE];
          idummy=std::max(1,numroc_(&b,&mb,&nodes[iend]->myrow,&IZERO,&nodes[iend]->nprow));
          descinit_(descRSr,&b,&d,&mb,&nb,&IZERO,&IZERO,&(nodes[iend]->context),&idummy,&ierr);
        } else if(!was_failed){
          /* Compress Rr_upd and Sc_upd, and initialize descRSc_upd */
          b=nodes[iend]->v_rank;
          tmp=new T[b*dinc];
          lacpy('N',b,dinc,Rr_upd,loc_htree.a_dim,tmp,b);
          delete[] Rr_upd;
          Rr_upd=tmp;

          tmp=new T[b*dinc];
          lacpy('N',b,dinc,Sc_upd,loc_htree.a_dim,tmp,b);
          delete[] Sc_upd;
          Sc_upd=tmp;

          descRSc_upd=new int[BLACSCTXTSIZE];
          idummy=std::max(1,numroc_(&b,&mb,&nodes[iend]->myrow,&IZERO,&nodes[iend]->nprow));
          descinit_(descRSc_upd,&b,&dinc,&mb,&nb,&IZERO,&IZERO,&(nodes[iend]->context),&idummy,&ierr);

          /* Compress Rc_upd and Sr_upd, and initialize descRSr_upd */
          b=nodes[iend]->u_rank;
          tmp=new T[b*dinc];
          lacpy('N',b,dinc,Rc_upd,loc_htree.a_dim,tmp,b);
          delete[] Rc_upd;
          Rc_upd=tmp;

          tmp=new T[b*dinc];
          lacpy('N',b,dinc,Sr_upd,loc_htree.a_dim,tmp,b);
          delete[] Sr_upd;
          Sr_upd=tmp;

          descRSr_upd=new int[BLACSCTXTSIZE];
          idummy=std::max(1,numroc_(&b,&mb,&nodes[iend]->myrow,&IZERO,&nodes[iend]->nprow));
          descinit_(descRSr_upd,&b,&dinc,&mb,&nb,&IZERO,&IZERO,&(nodes[iend]->context),&idummy,&ierr);

          if(nodes[iend]->parent->state==UNTOUCHED) {
            /* Merge vectors and updates; cf comment for interior nodes */
            tmpinfo.ctxt=nodes[iend]->context;
            tmpinfo.nprow=nodes[iend]->nprow;
            tmpinfo.npcol=nodes[iend]->npcol;
            tmpinfo.myrow=nodes[iend]->myrow;
            tmpinfo.mycol=nodes[iend]->mycol;
            tmpinfo.mb=ginfo->mb;
            tmpinfo.nb=ginfo->nb;
            updateColumns(&Sr,nodes[iend]->u_rank,d-dinc,&descRSr,Sr_upd,dinc,descRSr_upd,&tmpinfo);
            delete[] Sr_upd;
            updateColumns(&Sc,nodes[iend]->v_rank,d-dinc,&descRSc,Sc_upd,dinc,descRSc_upd,&tmpinfo);
            delete[] Sc_upd;
            updateColumns(&Rr,nodes[iend]->v_rank,d-dinc,&descRSc,Rr_upd,dinc,descRSc_upd,&tmpinfo);
            delete[] Rr_upd;
            updateColumns(&Rc,nodes[iend]->u_rank,d-dinc,&descRSr,Rc_upd,dinc,descRSr_upd,&tmpinfo);
            delete[] Rc_upd;
          }
        }
      }
    }
  }

  /* Third, parallel nodes are visited. This is the same algorithm as
   * in the lower part, except that arrays are constantly distributed
   * in a 2D block-cyclic format and parallel Interpolative Decomposition
   * is used.
   *
   * We separate leaves and interior nodes (interior nodes are found
   * in par_nodes).
   */

  /* Leaves */
  for(i=0;i<dim_tree;i++) {
    proc_b=nodes[i]->procs[0];
    proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;
    proc_e=nodes[i]->procs[1];
    if(proc_e>proc_b && nodes[i]->nr_children==0) {
      /* Leaf mapped on several processes */
      if(proc_g>=myid && proc_b<=myid) {
        /* myid is in the grid */

        /* a/ Extract a local sample
         * Sr_loc = Sr(ib:ie,:)-A(ib:ie,ib:ie) *Rr(ib:ie,:)
         * Sc_loc = Sc(ib:ie,:)-A(ib:ie,ib:ie)'*Rc(ib:ie,:)
         * with ib,ie the bounds of the clusters at leaf level.
         * Sr_loc and Sc_loc overwrite Sr(ib:ie,:) and Sc(ib:ie,:)
         *
         * The diagonal blocks is already in 2D block cyclic format
         * in "D".
         * At this point, X=Rr/Rc/Sr/Sc contains only X(ib:ie,:).
         */
        ib=nodes[i]->tree_ib-1;
        ie=nodes[i]->tree_ie-1;
        b=ie-ib+1;

        if(dosync)
          participate_to_calls(elements,nodes[i]->nb_calls_prev,comm);
        if(nodes[i]->state==UNTOUCHED) {
          if(elements==NULL) {
            /* Explicit matrix */
            nodes[i]->D=D;
          } else {
            /* User-given element access */
            locr=numroc_(&b,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
            locc=numroc_(&b,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
            idummy=std::max(1,locr);
            nodes[i]->D=new T[locr*locc];
            descD=new int[BLACSCTXTSIZE];
            descinit_(descD,&b,&b,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
            tmpI=new int[b];
            for(k=0;k<b;k++)
              tmpI[k]=nodes[i]->tree_ib+k;
            tmpJ=tmpI;
            if(dosync) {
              /* Send indices and descriptor to all the other processes
               * as they need them to participate to the collective call
               * to the user-given access routine.
               * proc_b is in charge of sending to processes that are
               * either 1/ mapped on i but out of the grid, or 2/ mapped
               * on other nodes, i.e. comm \ [proc_b...proc_g] */
              if(myid==proc_b) {
                for(k=0;k<np;k++) {
                  if(k<proc_b || k>proc_g) {
#if defined(debug)
                    std::cout << "dosync: " << proc_b << " sending information to " << k << std::endl;
#endif
                    MPI_Isend((void*)descD,BLACSCTXTSIZE,MPI_INTEGER,k,2,comm,&request);
                    MPI_Isend((void*)tmpI,b,MPI_INTEGER,k,2,comm,&request);
                    MPI_Isend((void*)tmpJ,b,MPI_INTEGER,k,2,comm,&request);
                  }
                }
              }
            }
            elements(obj,tmpI,tmpJ,nodes[i]->D,descD);
            if(dosync)
              MPI_Barrier(comm);
            delete[] tmpI;
          }
        }
        if(dosync)
          participate_to_calls(elements,nodes[i]->nb_calls_next,comm);

        if(i==dim_tree-1) {
          /* Nothing else to do for the root node */
          nodes[i]->state=COMPRESSED;
          break;
        }

        if(nodes[i]->state==UNTOUCHED) {
          /* UNTOUCHED: compute S-D*R */
          locrRS=numroc_(&b,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          loccRS=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          pgemm<T>('N','N',b,d,b,NONE,nodes[i]->D,IONE,IONE,descD,Rr,IONE,IONE,descRS,ONE,Sr,IONE,IONE,descRS);
          pgemm<T>('C','N',b,d,b,NONE,nodes[i]->D,IONE,IONE,descD,Rc,IONE,IONE,descRS,ONE,Sc,IONE,IONE,descRS);
        } else {
          /* PARTIALLY_COMPRESSED or COMPRESSED: computes S_upd-D*R_upd separately */
          locrRS=numroc_(&b,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          loccRS=numroc_(&dinc,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          pgemm<T>('N','N',b,dinc,b,NONE,nodes[i]->D,IONE,IONE,descD,Rr_upd,IONE,IONE,descRS_upd,ONE,Sr_upd,IONE,IONE,descRS_upd);
          pgemm<T>('C','N',b,dinc,b,NONE,nodes[i]->D,IONE,IONE,descD,Rc_upd,IONE,IONE,descRS_upd,ONE,Sc_upd,IONE,IONE,descRS_upd);
        }

        /* b/ Compress with Interpolative Decomposition: define the generators U and V
         * and the compressed indices Ir and Ic (resp. row and column indices).
         * Indices in output of ID are local to the matrix passed to the routine;
         * they are shifted in c/
         *
         * Sr/Sc are transposed into tmp before compression.
         *
         * After compression, Sr/Sc=Sr/Sc(Ir/Ic,:).
         */
        tmpinfo.ctxt=nodes[i]->context;
        tmpinfo.nprow=nodes[i]->nprow;
        tmpinfo.npcol=nodes[i]->npcol;
        tmpinfo.myrow=nodes[i]->myrow;
        tmpinfo.mycol=nodes[i]->mycol;
        tmpinfo.mb=ginfo->mb;
        tmpinfo.nb=ginfo->nb;

        if(nodes[i]->state==PARTIALLY_COMPRESSED) {
          /* Merge vectors and updates */
          updateColumns(&Sr,b,d-dinc,&descRS,Sr_upd,dinc,descRS_upd,&tmpinfo);
          delete[] Sr_upd;
          updateColumns(&Sc,b,d-dinc,&descRS,Sc_upd,dinc,descRS_upd,&tmpinfo);
          delete[] Sc_upd;
          updateColumns(&Rr,b,d-dinc,&descRS,Rr_upd,dinc,descRS_upd,&tmpinfo);
          delete[] Rr_upd;
          updateColumns(&Rc,b,d-dinc,&descRS,Rc_upd,dinc,descRS_upd,&tmpinfo);
          delete[] Rc_upd;
        }

        if(nodes[i]->state!=COMPRESSED) {
          /* Compression */
          locr=numroc_(&d,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&b,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          tmp=new T[locr*locc];
          idummy=std::max(locr,1);
          desc=new int[BLACSCTXTSIZE];
          descinit_(desc,&d,&b,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
          ptranc<T>(d,b,ONE,Sr,IONE,IONE,descRS,ZERO,tmp,IONE,IONE,desc);
          t=MPI_Wtime();
          par_XInterpol<T,S>(d,b,tmp,desc,tol,&nodes[i]->u_rank,&nodes[i]->U_E,&nodes[i]->Ir,&nodes[i]->U_perm,&tmpinfo);
          time_ID+=MPI_Wtime()-t;
          delete[] tmp;
          if(d-nodes[i]->u_rank<dlim && d+dinc<=dmax) {
            /* Reject this compression; go all the way to the root node
             * to get a global synchronization and restart the compression.
             */
            nodes[i]->state=PARTIALLY_COMPRESSED;
            if(myid==proc_b) {
              /* Notify out-of-grid processes */
              for(k=proc_g+1;k<=proc_e;k++)
                MPI_Send((void*) &nodes[i]->state,IONE,MPI_INTEGER,k,1,comm);
            }
            /* Go to the parent node */
            continue;
          }

          locr=numroc_(&d,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&b,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          tmp=new T[locr*locc];
          idummy=std::max(locr,1);
          desc=new int[BLACSCTXTSIZE];
          descinit_(desc,&d,&b,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
          ptranc<T>(d,b,ONE,Sc,IONE,IONE,descRS,ZERO,tmp,IONE,IONE,desc);
          t=MPI_Wtime();
          par_XInterpol<T,S>(d,b,tmp,desc,tol,&nodes[i]->v_rank,&nodes[i]->V_E,&nodes[i]->Ic,&nodes[i]->V_perm,&tmpinfo);
          time_ID+=MPI_Wtime()-t;
          delete[] tmp;
          if(d-nodes[i]->v_rank<dlim && d+dinc<=dmax) {
            /* Reject this compression; go all the way to the root node
             * to get a global synchronization and restart the compression.
             */
            nodes[i]->state=PARTIALLY_COMPRESSED;
            if(myid==proc_b) {
              /* Notify out-of-grid processes */
              for(k=proc_g+1;k<=proc_e;k++)
                MPI_Send((void*) &nodes[i]->state,IONE,MPI_INTEGER,k,1,comm);
            }
            /* Go to the parent node */
            continue;
          }

          /* Transposition */
          *r=std::max(*r,nodes[i]->u_rank);
          nodes[i]->u_rows=b-nodes[i]->u_rank;
          nodes[i]->descU=new int[BLACSCTXTSIZE];
          idummy=std::max(1,numroc_(&nodes[i]->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
          descinit_(nodes[i]->descU,&nodes[i]->u_rank,&nodes[i]->u_rows,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
          locr=numroc_(&nodes[i]->u_rows,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&nodes[i]->u_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          tmp= new T[locr*locc];
          idummy=std::max(locr,1);
          descinit_(desc,&nodes[i]->u_rows,&nodes[i]->u_rank,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
          ptranc<T>(nodes[i]->u_rows,nodes[i]->u_rank,ONE,nodes[i]->U_E,IONE,IONE,nodes[i]->descU,ZERO,tmp,IONE,IONE,desc);
          delete[] nodes[i]->U_E;
          nodes[i]->U_E=tmp;
          idummy=std::max(1,numroc_(&nodes[i]->u_rows,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
          descinit_(nodes[i]->descU,&nodes[i]->u_rows,&nodes[i]->u_rank,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);

          *r=std::max(*r,nodes[i]->v_rank);
          nodes[i]->v_rows=b-nodes[i]->v_rank;
          nodes[i]->descV=new int[BLACSCTXTSIZE];
          idummy=std::max(1,numroc_(&nodes[i]->v_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
          descinit_(nodes[i]->descV,&nodes[i]->v_rank,&b,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
          locr=numroc_(&nodes[i]->v_rows,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&nodes[i]->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          tmp= new T[locr*locc];
          idummy=std::max(locr,1);
          descinit_(desc,&nodes[i]->v_rows,&nodes[i]->v_rank,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
          ptranc<T>(nodes[i]->v_rows,nodes[i]->v_rank,ONE,nodes[i]->V_E,IONE,IONE,nodes[i]->descV,ZERO,tmp,IONE,IONE,desc);
          delete[] nodes[i]->V_E;
          nodes[i]->V_E=tmp;
          idummy=std::max(1,numroc_(&nodes[i]->v_rows,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
          descinit_(nodes[i]->descV,&nodes[i]->v_rows,&nodes[i]->v_rank,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);

          /* Extraction */
          parExtractSubmatrix<T,S>(Sr,b,d,descRS,nodes[i]->Ir,nodes[i]->u_rank,NULL,-1,&tmp,&tmpinfo);
          delete[] Sr;
          Sr=tmp;
          nodes[i]->Jr=new int[nodes[i]->u_rank];
          for(j=0;j<nodes[i]->u_rank;j++)
            nodes[i]->Jr[j]=nodes[i]->Ir[j];

          parExtractSubmatrix<T,S>(Sc,b,d,descRS,nodes[i]->Ic,nodes[i]->v_rank,NULL,-1,&tmp,&tmpinfo);
          delete[] Sc;
          Sc=tmp;
          nodes[i]->Jc=new int[nodes[i]->v_rank];
          for(j=0;j<nodes[i]->v_rank;j++)
            nodes[i]->Jc[j]=nodes[i]->Ic[j];
        } else {
          /* An ancestor of i restarted the traversal. We update the new columns of Sr/Sc; we
           * have to be careful because Ir/Ic have been previously transformed to absolute indices.
           * We use Jr/Jc instead. */
          parExtractSubmatrix<T,S>(Sr_upd,b,dinc,descRS_upd,nodes[i]->Jr,nodes[i]->u_rank,NULL,-1,&tmp,&tmpinfo);
          delete[] Sr_upd;
          Sr_upd=tmp;

          parExtractSubmatrix<T,S>(Sc_upd,b,dinc,descRS_upd,nodes[i]->Jc,nodes[i]->v_rank,NULL,-1,&tmp,&tmpinfo);
          delete[] Sc_upd;
          Sc_upd=tmp;

        }

        /* c/ Update random vectors and shift indices
         * Rr_tau=V'*Rr(ib:ie,:)
         * Rc_tau=U'*Rc(ib:ie,:)
         * Ir_tau=Ir_tau(Jr) (with Jr stored in Ir_tau already)
         * Ic_tau=Ic_tau(Jc) (with Jc stored in Ic_tau already)
         * Rr_tau and Rc_tau overwrite Rr and Rc
         *
         * We need new descriptors because the size change, and
         * because Rr/Sc and Rc/Sr now have different sizes
         * (resp v_rank x d and u_rank x d).
         */
        if(nodes[i]->state!=COMPRESSED) {
          locrRS=numroc_(&b,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          loccRS=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          tmp=new T[locrRS*loccRS];
          lacpy('N',locrRS,loccRS,Rr,locrRS,tmp,locrRS);
          descRSc=new int[BLACSCTXTSIZE];
          locr=numroc_(&nodes[i]->v_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          idummy=std::max(1,locr);
          descinit_(descRSc,&nodes[i]->v_rank,&d,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
          delete[] Rr;
          Rr=new T[locr*locc];
          nodes[i]->par_applyGenerator('C','V',d,ONE,tmp,descRS,ZERO,Rr,descRSc,&tmpinfo);
          for(j=0;j<nodes[i]->u_rank;j++)
            nodes[i]->Ir[j]+=ib;
          delete[] tmp;

          tmp=new T[locrRS*loccRS];
          lacpy('N',locrRS,loccRS,Rc,locrRS,tmp,locrRS);
          descRSr=new int[BLACSCTXTSIZE];
          locr=numroc_(&nodes[i]->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          idummy=std::max(1,locr);
          descinit_(descRSr,&nodes[i]->u_rank,&d,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
          delete[] Rc;
          Rc=new T[locr*locc];
          nodes[i]->par_applyGenerator('C','U',d,ONE,tmp,descRS,ZERO,Rc,descRSr,&tmpinfo);
          for(j=0;j<nodes[i]->v_rank;j++)
            nodes[i]->Ic[j]+=ib;
          delete[] tmp;
        } else {
          /* For previously compressed nodes, only touch the update. */
          locrRS=numroc_(&b,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          loccRS=numroc_(&dinc,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          tmp=new T[locrRS*loccRS];
          lacpy('N',locrRS,loccRS,Rr_upd,locrRS,tmp,locrRS);
          descRSc_upd=new int[BLACSCTXTSIZE];
          locr=numroc_(&nodes[i]->v_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&dinc,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          idummy=std::max(1,locr);
          descinit_(descRSc_upd,&nodes[i]->v_rank,&dinc,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
          delete[] Rr_upd;
          Rr_upd=new T[locr*locc];
          nodes[i]->par_applyGenerator('C','V',dinc,ONE,tmp,descRS_upd,ZERO,Rr_upd,descRSc_upd,&tmpinfo);
          delete[] tmp;

          tmp=new T[locrRS*loccRS];
          lacpy('N',locrRS,loccRS,Rc_upd,locrRS,tmp,locrRS);
          descRSr_upd=new int[BLACSCTXTSIZE];
          locr=numroc_(&nodes[i]->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&dinc,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          idummy=std::max(1,locr);
          descinit_(descRSr_upd,&nodes[i]->u_rank,&dinc,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
          delete[] Rc_upd;
          Rc_upd=new T[locr*locc];
          nodes[i]->par_applyGenerator('C','U',dinc,ONE,tmp,descRS_upd,ZERO,Rc_upd,descRSr_upd,&tmpinfo);
          delete[] tmp;
        }


        if(nodes[i]->state==COMPRESSED && nodes[i]->parent->state==UNTOUCHED) {
          /* Merge vectors and updates */
          updateColumns(&Sr,nodes[i]->u_rank,d-dinc,&descRSr,Sr_upd,dinc,descRSr_upd,&tmpinfo);
          delete[] Sr_upd;
          updateColumns(&Sc,nodes[i]->v_rank,d-dinc,&descRSc,Sc_upd,dinc,descRSc_upd,&tmpinfo);
          delete[] Sc_upd;
          updateColumns(&Rr,nodes[i]->v_rank,d-dinc,&descRSc,Rr_upd,dinc,descRSc_upd,&tmpinfo);
          delete[] Rr_upd;
          updateColumns(&Rc,nodes[i]->u_rank,d-dinc,&descRSr,Rc_upd,dinc,descRSr_upd,&tmpinfo);
          delete[] Rc_upd;
        }

        /* Mark node as compressed */
        nodes[i]->numrand=d;
        nodes[i]->state=COMPRESSED;

        /* Processes out of the grid for the current node might be active at other nodes.
         * They need to know the state, and ranks and indices if COMPRESSED
         * (otherwise they simply go to the parent node).
         * They receive the info from proc_b.
         */
        if(myid==proc_b) {
          for(k=proc_g+1;k<=proc_e;k++)
            MPI_Send((void*) &nodes[i]->state,IONE,MPI_INTEGER,k,1,comm);

          for(k=proc_g+1;k<=proc_e;k++)
            MPI_Send((void*) &nodes[i]->u_rank,IONE,MPI_INTEGER,k,1,comm);
          for(k=proc_g+1;k<=proc_e;k++)
            MPI_Send((void*) &nodes[i]->u_rows,IONE,MPI_INTEGER,k,1,comm);
          if(nodes[i]->u_rank>0)
            for(k=proc_g+1;k<=proc_e;k++)
              MPI_Send((void*) nodes[i]->Ir,nodes[i]->u_rank,MPI_INTEGER,k,1,comm);

          for(k=proc_g+1;k<=proc_e;k++)
            MPI_Send((void*) &nodes[i]->v_rank,IONE,MPI_INTEGER,k,1,comm);
          for(k=proc_g+1;k<=proc_e;k++)
            MPI_Send((void*) &nodes[i]->v_rows,IONE,MPI_INTEGER,k,1,comm);
          if(nodes[i]->v_rank>0)
            for(k=proc_g+1;k<=proc_e;k++)
              MPI_Send((void*) nodes[i]->Ic,nodes[i]->v_rank,MPI_INTEGER,k,1,comm);
        }

      } else if(myid<=proc_e && myid>=proc_g+1) {
        /* Processes out of the grid also participate to collective
         * calls to the user-given access routine.
         */
        if(dosync)
          participate_to_calls(elements,nb_calls_level[nodes[i]->level],comm);

        /* Processes on which i is mapped, but out of the grid */
        if(i==dim_tree-1) {
          nodes[i]->state=COMPRESSED;
          break;
        }

        /* Processes out of the grid for the current might be active at other nodes.
         * They need to know the state, and ranks and indices if COMPRESSED
         * (otherwise they simply go to the parent node).
         * They receive the info from proc_b.
         */
        MPI_Recv((void*) &nodes[i]->state,IONE,MPI_INTEGER,proc_b,1,comm,&status);
        if(nodes[i]->state==COMPRESSED) {
          nodes[i]->numrand=d;
          MPI_Recv((void*) &nodes[i]->u_rank,IONE,MPI_INTEGER,proc_b,1,comm,&status);
          MPI_Recv((void*) &nodes[i]->u_rows,IONE,MPI_INTEGER,proc_b,1,comm,&status);
          if(nodes[i]->u_rank>0) {
            nodes[i]->Ir=new int[nodes[i]->u_rank];
            MPI_Recv((void*) nodes[i]->Ir,nodes[i]->u_rank,MPI_INTEGER,proc_b,1,comm,&status);
          }

          MPI_Recv((void*) &nodes[i]->v_rank,IONE,MPI_INTEGER,proc_b,1,comm,&status);
          MPI_Recv((void*) &nodes[i]->v_rows,IONE,MPI_INTEGER,proc_b,1,comm,&status);
          if(nodes[i]->v_rank>0) {
            nodes[i]->Ic=new int[nodes[i]->v_rank];
            MPI_Recv((void*) nodes[i]->Ic,nodes[i]->v_rank,MPI_INTEGER,proc_b,1,comm,&status);
          }
        }
      }
    }
  }

  /* Interior nodes */
  for(j=0;j<par_lvl;j++) {
    /* Adaptive rank mechanism: dealing with R/S and R/S_upd.
     * If a child is COMPRESSED (thus has _upd vectors) and its
     * parent is UNTOUCHED, then the R/S_upd are appended to R/S
     * at the child node. Then, scenarios at the parent node:
     *  - If the node is compressed, then we are in a restart phase, but the
     *    node and all its descendants have been previously successfully
     *    compressed. The children provide _upd vectors and the node computes
     *    updates on these vectors and pass them to its parents.
     *  - If the node is partially compressed, then we are in a restart phase,
     *    but, necessarily, all its descendants have been successfully
     *    compressed (during a previous traversal). The children provide _upd
     *    vectors, and the node append them to R/S to perform compression. Then
     *    the node passes R/S (no more _upd) to its parent.
     *  - If a node is untouched, its children can be either:
     *    - Compressed from a previous traversal: as indicated above,
     *    they merge R/S_upd into R/S and only provide R/S to the parent.
     *    - Compressed during the current traversal: they just computed
     *    R/S and provide them to the parent.
     */
    i=par_nodes[j];
    ib=nodes[i]->tree_ib-1;
    ie=nodes[i]->tree_ie-1;
    child1=nodes[i]->children[0];
    child2=nodes[i]->children[1];

    /* Processes exchange child states */
    exchangeInteger(&child1->state,child1->procs[0],child1->procs[1],&child2->state,child2->procs[0],child2->procs[1],comm);
    /* If not both children are compressed, go to the next node.
     * Processes go through all their ancestors without doing anything
     * (besides informing the processes working at siblings).
     * Eventually, all the processes reach the root node and restart.
     */
    if(child1->state!=COMPRESSED || child2->state!=COMPRESSED)
      continue;

    /* Node i is mapped on proc_b...proc_e but only
     * proc_b...proc_g are active (in the grid).
     */
    proc_b=nodes[i]->procs[0];
    proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;
    proc_e=nodes[i]->procs[1];
    if(ib!=child1->tree_ib-1 || ie!=child2->tree_ie-1 || child1->tree_ie+1!=child2->tree_ib) {
      std::cout << "Problem with clustering indices." << std::endl;
      exit(1);
    }
    if(nodes[i]->state==UNTOUCHED) {
      exchangeIndices(&child1->Ir,&child1->u_rank,child1->procs[0],child1->procs[1],&child2->Ir,&child2->u_rank,child2->procs[0],child2->procs[1],comm);
      exchangeIndices(&child1->Ic,&child1->v_rank,child1->procs[0],child1->procs[1],&child2->Ic,&child2->v_rank,child2->procs[0],child2->procs[1],comm);
      /* Processes exchange u_rows and v_rows; this is useful in the factorization
       * and solution phases, but not needed here.
       */
      exchangeInteger(&child1->u_rows,child1->procs[0],child1->procs[1],&child2->u_rows,child2->procs[0],child2->procs[1],comm);
      exchangeInteger(&child1->v_rows,child1->procs[0],child1->procs[1],&child2->v_rows,child2->procs[0],child2->procs[1],comm);
      if(child1->u_rank<0 || child2->u_rank<0 || child1->v_rank<0 || child2->v_rank<0) {
        std::cout << "Internal error 2 in HSSR_par_Compress" << std::endl;
        exit(-1);
      }
    }
    br=child1->u_rank+child2->u_rank;
    bc=child1->v_rank+child2->v_rank;
    tmpinfo.ctxt=nodes[i]->context;
    tmpinfo.nprow=nodes[i]->nprow;
    tmpinfo.npcol=nodes[i]->npcol;
    tmpinfo.myrow=nodes[i]->myrow;
    tmpinfo.mycol=nodes[i]->mycol;
    tmpinfo.mb=ginfo->mb;
    tmpinfo.nb=ginfo->nb;

    /* Redistributing the random vectors and samples.
     * NB: Rr and Sc have bc rows, while Rc and Sr have br rows (!)
     */
    if(i!=dim_tree-1) {
      if(nodes[i]->state==UNTOUCHED) {
        mergeVector(Rr,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,child1->v_rank,IONE,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,child2->v_rank,IONE,proc_b,proc_g,d,descRSc,&tmp,&desc,nodes[i]->allctxt,&tmpinfo,comm);
        delete[] desc;
        delete[] Rr;
        Rr=tmp;
        mergeVector(Sr,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,child1->u_rank,IONE,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,child2->u_rank,IONE,proc_b,proc_g,d,descRSr,&tmp,&desc,nodes[i]->allctxt,&tmpinfo,comm);
        delete[] desc;
        delete[] Sr;
        Sr=tmp;
        mergeVector(Rc,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,child1->u_rank,IONE,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,child2->u_rank,IONE,proc_b,proc_g,d,descRSr,&tmp,&desc,nodes[i]->allctxt,&tmpinfo,comm);
        delete[] Rc;
        Rc=tmp;
        delete[] descRSr;
        descRSr=desc;
        mergeVector(Sc,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,child1->v_rank,IONE,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,child2->v_rank,IONE,proc_b,proc_g,d,descRSc,&tmp,&desc,nodes[i]->allctxt,&tmpinfo,comm);
        delete[] Sc;
        Sc=tmp;
        delete[] descRSc;
        descRSc=desc;
      } else {
        mergeVector(Rr_upd,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,child1->v_rank,IONE,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,child2->v_rank,IONE,proc_b,proc_g,dinc,descRSc_upd,&tmp,&desc,nodes[i]->allctxt,&tmpinfo,comm);
        delete[] desc;
        delete[] Rr_upd;
        Rr_upd=tmp;
        mergeVector(Sr_upd,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,child1->u_rank,IONE,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,child2->u_rank,IONE,proc_b,proc_g,dinc,descRSr_upd,&tmp,&desc,nodes[i]->allctxt,&tmpinfo,comm);
        delete[] desc;
        delete[] Sr_upd;
        Sr_upd=tmp;
        mergeVector(Rc_upd,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,child1->u_rank,IONE,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,child2->u_rank,IONE,proc_b,proc_g,dinc,descRSr_upd,&tmp,&desc,nodes[i]->allctxt,&tmpinfo,comm);
        delete[] Rc_upd;
        Rc_upd=tmp;
        delete[] descRSr_upd;
        descRSr_upd=desc;
        mergeVector(Sc_upd,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,child1->v_rank,IONE,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,child2->v_rank,IONE,proc_b,proc_g,dinc,descRSc_upd,&tmp,&desc,nodes[i]->allctxt,&tmpinfo,comm);
        delete[] Sc_upd;
        Sc_upd=tmp;
        delete[] descRSc_upd;
        descRSc_upd=desc;
      }
    }

    /* From here, exclude processes that are not in the grid of i */
    if(myid>=proc_b && myid<=proc_g) {
      /* a/ Extract a local sample
       * Sr_loc= [Sr_1-B12 *Rr_2;Sr_2-B21 *Rr_1]
       * Sc_loc= [Sc_1-B21'*Rc_2;Sc_2-B12'*Rc_1]
       * with B12=A(Ir_1,Ic_2) and B21=A(Ir_2,Ic_1)
       * with _1 and _2 the children of i.
       * Sr_loc and Sc_loc overwrite Sr and Sc.
       *
       * Before this, A12[j] and A21[j] contain all the values of the
       * corresponding block of the original matrix, not B12=A(Ir_1,Ic_2) and
       * B21=A(Ir_2,Ic_1). The first step is to extract the needed values into
       * B12/B21 and free A12[j]/A21[j] that are no longer needed. For this,
       * the ranks and indices must be exchanged. Also, indices need to be
       * temporarily translated to find the right entries in A12/A21.
       *
       * Then, for the samples: call procs_i the set of processes working at i
       * and procs_c1 (procs_c2) the set of processes working at its child c1
       * (c2). procs_i=proc_c1 U procs_c2.
       * Processes procs_i own B12 and B21.
       * Processes procs_c1 own Sr/Sc/Rr/Rc_1.
       * Processes procs_c2 own Sr/Sc/Rr/Rc_2.
       * Sr/c_loc:
       *  1/ Processes redistribute Sr/Sc/Rr/Rc_1/2 from procs_c1/2 to
       *  everybody (procs_i) into tmp
       *  2/ tmp erases Sr/Sc/Rr/Rc_1/2.
       *  3/ Everybody owns what is need for the products that define Sr/c_loc;
       *  do 2 pgemms. The result erases Sr/Sc.
       */

      /* Extracting B12 and B21. */
      if(dosync)
        participate_to_calls(elements,nodes[i]->nb_calls_prev,comm);
      if(nodes[i]->state==UNTOUCHED) {
        if(elements==NULL) {
          /* Explicit matrix.
           * Shift global (1..n) indices to local (1...size A12/A21)
           * before extracting, and shift back after.
           */
          for(k=0;k<child1->u_rank;k++)
            child1->Ir[k]-=child1->tree_ib-1;
          for(k=0;k<child2->v_rank;k++)
            child2->Ic[k]-=child2->tree_ib-1;
          idummy1=child1->tree_ie-child1->tree_ib+1;
          idummy2=child2->tree_ie-child2->tree_ib+1;
          idummy=std::max(1,numroc_(&idummy1,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
          desc=new int[BLACSCTXTSIZE];
          descinit_(desc,&idummy1,&idummy2,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
          parExtractSubmatrix<T,S>(A12[j],idummy1,idummy2,desc,child1->Ir,child1->u_rank,child2->Ic,child2->v_rank,&nodes[i]->B12,&tmpinfo);
          delete[] A12[j];
          for(k=0;k<child1->u_rank;k++)
            child1->Ir[k]+=child1->tree_ib-1;
          for(k=0;k<child2->v_rank;k++)
            child2->Ic[k]+=child2->tree_ib-1;

          /* Shift global (1..n) indices to local (1...size A12/A21)
           * before extracting, and shift back after.
           */
          for(k=0;k<child2->u_rank;k++)
            child2->Ir[k]-=child2->tree_ib-1;
          for(k=0;k<child1->v_rank;k++)
            child1->Ic[k]-=child1->tree_ib-1;
          idummy1=child2->tree_ie-child2->tree_ib+1;
          idummy2=child1->tree_ie-child1->tree_ib+1;
          idummy=std::max(1,numroc_(&idummy1,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
          descinit_(desc,&idummy1,&idummy2,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
          parExtractSubmatrix<T,S>(A21[j],idummy1,idummy2,desc,child2->Ir,child2->u_rank,child1->Ic,child1->v_rank,&nodes[i]->B21,&tmpinfo);
          delete[] A21[j];
          for(k=0;k<child2->u_rank;k++)
            child2->Ir[k]+=child2->tree_ib-1;
          for(k=0;k<child1->v_rank;k++)
            child1->Ic[k]+=child1->tree_ib-1;
        } else {
          /* User-given element access */
          locr=numroc_(&child1->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&child2->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          idummy=std::max(1,locr);
          desc=new int[BLACSCTXTSIZE];
          descinit_(desc,&child1->u_rank,&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
          nodes[i]->B12=new T[locr*locc];
          if(dosync) {
            if(myid==proc_b) {
              for(k=0;k<np;k++) {
                if(k<proc_b || k>proc_g) {
#if defined(debug)
                  std::cout << "dosync: " << proc_b << " sending information to " << k << std::endl;
#endif
                  MPI_Isend((void*)desc,BLACSCTXTSIZE,MPI_INTEGER,k,2,comm,&request);
                  MPI_Isend((void*)child1->Ir,child1->u_rank,MPI_INTEGER,k,2,comm,&request);
                  MPI_Isend((void*)child2->Ic,child2->v_rank,MPI_INTEGER,k,2,comm,&request);
                }
              }
            }
          }
          elements(obj,child1->Ir,child2->Ic,nodes[i]->B12,desc);
          if(dosync)
            MPI_Barrier(comm);

          locr=numroc_(&child2->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&child1->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          idummy=std::max(1,locr);
          descinit_(desc,&child2->u_rank,&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
          nodes[i]->B21=new T[locr*locc];
          if(dosync) {
            if(myid==proc_b) {
              for(k=0;k<np;k++) {
                if(k<proc_b || k>proc_g) {
#if defined(debug)
                  std::cout << "dosync: " << proc_b << " sending information to " << k << std::endl;
#endif
                  MPI_Isend((void*)desc,BLACSCTXTSIZE,MPI_INTEGER,k,2,comm,&request);
                  MPI_Isend((void*)child2->Ir,child2->u_rank,MPI_INTEGER,k,2,comm,&request);
                  MPI_Isend((void*)child1->Ic,child1->v_rank,MPI_INTEGER,k,2,comm,&request);
                }
              }
            }
          }
          elements(obj,child2->Ir,child1->Ic,nodes[i]->B21,desc);
          if(dosync)
            MPI_Barrier(comm);
        }
      }
      if(dosync)
        participate_to_calls(elements,nodes[i]->nb_calls_next,comm);

      /* If root node, nothing else to do */
      if(i==dim_tree-1) {
        nodes[i]->state=COMPRESSED;
        break;
      }

      /* Form the local sample (it erases the current sample) */
      if(nodes[i]->state==UNTOUCHED) {
        idummy=std::max(1,numroc_(&child1->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        desc=new int[BLACSCTXTSIZE];
        descinit_(desc,&child1->u_rank,&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        /* First rows: Sr_1-B12 *Rr_2; Rr_2 is in the last rows of Rr */
        pgemm<T>('N','N',child1->u_rank,d,child2->v_rank,NONE,nodes[i]->B12,IONE,IONE,desc,Rr,child1->v_rank+1,IONE,descRSc,ONE,Sr,IONE,IONE,descRSr);

        idummy=std::max(1,numroc_(&child2->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(desc,&child2->u_rank,&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        /* Last rows: Sr_2-B21 *Rr_1; Rr_1 is in the first rows of Rr */
        pgemm<T>('N','N',child2->u_rank,d,child1->v_rank,NONE,nodes[i]->B21,IONE,IONE,desc,Rr,IONE,IONE,descRSc,ONE,Sr,child1->u_rank+1,IONE,descRSr);

        idummy=std::max(1,numroc_(&child2->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(desc,&child2->u_rank,&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        /* First rows: Sc_1-B21' *Rc_2; Rc_2 is in the last rows of Rr */
        pgemm<T>('C','N',child1->v_rank,d,child2->u_rank,NONE,nodes[i]->B21,IONE,IONE,desc,Rc,child1->u_rank+1,IONE,descRSr,ONE,Sc,IONE,IONE,descRSc);

        idummy=std::max(1,numroc_(&child1->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(desc,&child1->u_rank,&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        /* Last rows: Sc_2-B12'*Rc_1; Rc_1 is in the first rows of Rr */
        pgemm<T>('C','N',child2->v_rank,d,child1->u_rank,NONE,nodes[i]->B12,IONE,IONE,desc,Rc,IONE,IONE,descRSr,ONE,Sc,child1->v_rank+1,IONE,descRSc);
      } else {
        idummy=std::max(1,numroc_(&child1->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        desc=new int[BLACSCTXTSIZE];
        descinit_(desc,&child1->u_rank,&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        /* First rows: Sr_1-B12 *Rr_2; Rr_2 is in the last rows of Rr */
        pgemm<T>('N','N',child1->u_rank,dinc,child2->v_rank,NONE,nodes[i]->B12,IONE,IONE,desc,Rr_upd,child1->v_rank+1,IONE,descRSc_upd,ONE,Sr_upd,IONE,IONE,descRSr_upd);

        idummy=std::max(1,numroc_(&child2->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(desc,&child2->u_rank,&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        /* Last rows: Sr_2-B21 *Rr_1; Rr_1 is in the first rows of Rr */
        pgemm<T>('N','N',child2->u_rank,dinc,child1->v_rank,NONE,nodes[i]->B21,IONE,IONE,desc,Rr_upd,IONE,IONE,descRSc_upd,ONE,Sr_upd,child1->u_rank+1,IONE,descRSr_upd);

        idummy=std::max(1,numroc_(&child2->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(desc,&child2->u_rank,&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        /* First rows: Sc_1-B21' *Rc_2; Rc_2 is in the last rows of Rr */
        pgemm<T>('C','N',child1->v_rank,dinc,child2->u_rank,NONE,nodes[i]->B21,IONE,IONE,desc,Rc_upd,child1->u_rank+1,IONE,descRSr_upd,ONE,Sc_upd,IONE,IONE,descRSc_upd);

        idummy=std::max(1,numroc_(&child1->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(desc,&child1->u_rank,&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        /* Last rows: Sc_2-B12'*Rc_1; Rc_1 is in the first rows of Rr */
        pgemm<T>('C','N',child2->v_rank,dinc,child1->u_rank,NONE,nodes[i]->B12,IONE,IONE,desc,Rc_upd,IONE,IONE,descRSr_upd,ONE,Sc_upd,child1->v_rank+1,IONE,descRSc_upd);
      }

      /* b/ Compress with Interpolative Decomposition: define the generators U and V
       * and the compressed indices Ir and Ic (resp. row and column indices).
       * U and V stay transposed for step c/
       * Indices in output of ID are local to the matrix passed to the routine;
       * they are shifted in c/
       *
       * Sr/Sc are transposed into tmp before compression.
       *
       * After compression, Sr/Sc=Sr/Sc(Ir/Ic,:).
       */
      tmpinfo.ctxt=nodes[i]->context;
      tmpinfo.nprow=nodes[i]->nprow;
      tmpinfo.npcol=nodes[i]->npcol;
      tmpinfo.myrow=nodes[i]->myrow;
      tmpinfo.mycol=nodes[i]->mycol;
      tmpinfo.mb=ginfo->mb;
      tmpinfo.nb=ginfo->nb;

      if(nodes[i]->state==PARTIALLY_COMPRESSED) {
        /* Merge vectors and updates */
        updateColumns(&Sr,br,d-dinc,&descRSr,Sr_upd,dinc,descRSr_upd,&tmpinfo);
        delete[] Sr_upd;
        updateColumns(&Sc,bc,d-dinc,&descRSc,Sc_upd,dinc,descRSc_upd,&tmpinfo);
        delete[] Sc_upd;
        updateColumns(&Rr,bc,d-dinc,&descRSc,Rr_upd,dinc,descRSc_upd,&tmpinfo);
        delete[] Rr_upd;
        updateColumns(&Rc,br,d-dinc,&descRSr,Rc_upd,dinc,descRSr_upd,&tmpinfo);
        delete[] Rc_upd;
      }

      if(nodes[i]->state!=COMPRESSED) {
        /* Compression */
        locr=numroc_(&d,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&br,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        tmp= new T[locr*locc];
        idummy=std::max(locr,1);
        desc=new int[BLACSCTXTSIZE];
        descinit_(desc,&d,&br,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        ptranc<T>(d,br,ONE,Sr,IONE,IONE,descRSr,ZERO,tmp,IONE,IONE,desc);
        t=MPI_Wtime();
        par_XInterpol<T,S>(d,br,tmp,desc,tol,&nodes[i]->u_rank,&nodes[i]->U_E,&nodes[i]->Ir,&nodes[i]->U_perm,&tmpinfo);
        time_ID+=MPI_Wtime()-t;
        delete[] tmp;
        if(d-nodes[i]->u_rank<dlim && d+dinc<=dmax) {
          /* Reject this compression; go all the way to the root node
           * to get a global synchronization and restart the compression.
           */
          nodes[i]->state=PARTIALLY_COMPRESSED;
          if(myid==proc_b) {
            /* Notify out-of-grid processes */
            for(k=proc_g+1;k<=proc_e;k++)
              MPI_Send((void*) &nodes[i]->state,IONE,MPI_INTEGER,k,1,comm);
          }
          /* Go to the parent node */
          continue;
        }

        locr=numroc_(&d,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&bc,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        tmp= new T[locr*locc];
        idummy=std::max(locr,1);
        descinit_(desc,&d,&bc,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        ptranc<T>(d,bc,ONE,Sc,IONE,IONE,descRSc,ZERO,tmp,IONE,IONE,desc);
        t=MPI_Wtime();
        par_XInterpol<T,S>(d,bc,tmp,desc,tol,&nodes[i]->v_rank,&nodes[i]->V_E,&nodes[i]->Ic,&nodes[i]->V_perm,&tmpinfo);
        time_ID+=MPI_Wtime()-t;
        delete[] tmp;
        if(d-nodes[i]->v_rank<dlim && d+dinc<=dmax) {
          /* Reject this compression; go all the way to the root node
           * to get a global synchronization and restart the compression.
           */
          nodes[i]->state=PARTIALLY_COMPRESSED;
          if(myid==proc_b) {
            /* Notify out-of-grid processes */
            for(k=proc_g+1;k<=proc_e;k++)
              MPI_Send((void*) &nodes[i]->state,IONE,MPI_INTEGER,k,1,comm);
          }
          /* Go to the parent node */
          continue;
        }

        /* Transposition */
        *r=std::max(*r,nodes[i]->u_rank);
        idummy=std::max(1,numroc_(&nodes[i]->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        nodes[i]->descU=new int[BLACSCTXTSIZE];
        nodes[i]->u_rows=br-nodes[i]->u_rank;
        descinit_(nodes[i]->descU,&nodes[i]->u_rank,&nodes[i]->u_rows,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        locr=numroc_(&nodes[i]->u_rows,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&nodes[i]->u_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        tmp= new T[locr*locc];
        idummy=std::max(locr,1);
        descinit_(desc,&nodes[i]->u_rows,&nodes[i]->u_rank,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        ptranc<T>(nodes[i]->u_rows,nodes[i]->u_rank,ONE,nodes[i]->U_E,IONE,IONE,nodes[i]->descU,ZERO,tmp,IONE,IONE,desc);
        delete[] nodes[i]->U_E;
        nodes[i]->U_E=tmp;
        idummy=std::max(1,numroc_(&nodes[i]->u_rows,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(nodes[i]->descU,&nodes[i]->u_rows,&nodes[i]->u_rank,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);

        *r=std::max(*r,nodes[i]->v_rank);
        idummy=std::max(1,numroc_(&nodes[i]->v_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        nodes[i]->descV=new int[BLACSCTXTSIZE];
        nodes[i]->v_rows=bc-nodes[i]->v_rank;
        descinit_(nodes[i]->descV,&nodes[i]->v_rank,&nodes[i]->v_rows,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        locr=numroc_(&nodes[i]->v_rows,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&nodes[i]->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        tmp= new T[locr*locc];
        idummy=std::max(locr,1);
        descinit_(desc,&nodes[i]->v_rows,&nodes[i]->v_rank,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        ptranc<T>(nodes[i]->v_rows,nodes[i]->v_rank,ONE,nodes[i]->V_E,IONE,IONE,nodes[i]->descV,ZERO,tmp,IONE,IONE,desc);
        delete[] nodes[i]->V_E;
        nodes[i]->V_E=tmp;
        idummy=std::max(1,numroc_(&nodes[i]->v_rows,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(nodes[i]->descV,&nodes[i]->v_rows,&nodes[i]->v_rank,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);

        /* Extraction */
        parExtractSubmatrix<T,S>(Sr,br,d,descRSr,nodes[i]->Ir,nodes[i]->u_rank,NULL,-1,&tmp,&tmpinfo);
        delete[] Sr;
        Sr=tmp;
        nodes[i]->Jr=new int[nodes[i]->u_rank];
        for(k=0;k<nodes[i]->u_rank;k++)
          nodes[i]->Jr[k]=nodes[i]->Ir[k];

        parExtractSubmatrix<T,S>(Sc,bc,d,descRSc,nodes[i]->Ic,nodes[i]->v_rank,NULL,-1,&tmp,&tmpinfo);
        delete[] Sc;
        Sc=tmp;
        nodes[i]->Jc=new int[nodes[i]->v_rank];
        for(k=0;k<nodes[i]->v_rank;k++)
          nodes[i]->Jc[k]=nodes[i]->Ic[k];
      } else {
        /* An ancestor of i restarted the traversal. We update the new columns of Sr/Sc; we
         * have to be careful because Ir/Ic have been previously transformed to absolute indices.
         * We use Jr/Jc instead. */
        parExtractSubmatrix<T,S>(Sr_upd,br,dinc,descRSr_upd,nodes[i]->Jr,nodes[i]->u_rank,NULL,-1,&tmp,&tmpinfo);
        delete[] Sr_upd;
        Sr_upd=tmp;

        parExtractSubmatrix<T,S>(Sc_upd,bc,dinc,descRSc_upd,nodes[i]->Jc,nodes[i]->v_rank,NULL,-1,&tmp,&tmpinfo);
        delete[] Sc_upd;
        Sc_upd=tmp;
      }

      /* c/ Update random vectors and shift indices
       * Rr_tau=V'*[Rr_1; Rr_2]
       * Rc_tau=U'*[Rc_1; Rc_2]
       * Ir_tau=[Ir_1 Ir_2](Jr)
       * Ic_tau=[Ic_1 Ic_2](Jc)
       *
       * Rr_1 and Rr_2 (resp. Rc_1 and Rc_2) are already stacked together
       * in Rr (resp. Rc). We make a copy in tmp before the pgemm. We need
       * to update the descriptors for Rr/c since the size changes.
       */
      if(nodes[i]->state!=COMPRESSED) {
        locr=numroc_(&bc,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        tmp=new T[locr*locc];
        placpy('N',bc,d,Rr,IONE,IONE,descRSc,tmp,IONE,IONE,descRSc);
        idummy=std::max(1,numroc_(&nodes[i]->v_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(desc,&nodes[i]->v_rank,&d,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        nodes[i]->par_applyGenerator('C','V',d,ONE,tmp,descRSc,ZERO,Rr,desc,&tmpinfo);
        descinit_(descRSc,&nodes[i]->v_rank,&d,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        delete[] tmp;
        for(k=0;k<nodes[i]->u_rank;k++)
          if(nodes[i]->Ir[k]<=child1->u_rank)
            nodes[i]->Ir[k]=child1->Ir[nodes[i]->Ir[k]-1];
          else
            nodes[i]->Ir[k]=child2->Ir[nodes[i]->Ir[k]-child1->u_rank-1];

        locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        tmp=new T[locr*locc];
        placpy('N',br,d,Rc,IONE,IONE,descRSr,tmp,IONE,IONE,descRSr);
        idummy=std::max(1,numroc_(&nodes[i]->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(desc,&nodes[i]->u_rank,&d,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        nodes[i]->par_applyGenerator('C','U',d,ONE,tmp,descRSr,ZERO,Rc,desc,&tmpinfo);
        descinit_(descRSr,&nodes[i]->u_rank,&d,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        delete[] tmp;
        for(k=0;k<nodes[i]->v_rank;k++)
          if(nodes[i]->Ic[k]<=child1->v_rank)
            nodes[i]->Ic[k]=child1->Ic[nodes[i]->Ic[k]-1];
          else
            nodes[i]->Ic[k]=child2->Ic[nodes[i]->Ic[k]-child1->v_rank-1];
      } else {
        /* For previously compressed nodes, only touch the update. */
        locr=numroc_(&bc,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&dinc,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        tmp=new T[locr*locc];
        placpy('N',bc,dinc,Rr_upd,IONE,IONE,descRSc_upd,tmp,IONE,IONE,descRSc_upd);
        idummy=std::max(1,numroc_(&nodes[i]->v_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(desc,&nodes[i]->v_rank,&dinc,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        nodes[i]->par_applyGenerator('C','V',dinc,ONE,tmp,descRSc_upd,ZERO,Rr_upd,desc,&tmpinfo);
        descinit_(descRSc_upd,&nodes[i]->v_rank,&dinc,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        delete[] tmp;

        locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&dinc,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        tmp=new T[locr*locc];
        placpy('N',br,dinc,Rc_upd,IONE,IONE,descRSr_upd,tmp,IONE,IONE,descRSr_upd);
        idummy=std::max(1,numroc_(&nodes[i]->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(desc,&nodes[i]->u_rank,&dinc,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        nodes[i]->par_applyGenerator('C','U',dinc,ONE,tmp,descRSr_upd,ZERO,Rc_upd,desc,&tmpinfo);
        descinit_(descRSr_upd,&nodes[i]->u_rank,&dinc,&mb,&nb,&IZERO,&IZERO,&(nodes[i]->context),&idummy,&ierr);
        delete[] tmp;
      }

      if(nodes[i]->state==COMPRESSED && nodes[i]->parent->state==UNTOUCHED) {
        /* Merge vectors and updates */
        updateColumns(&Sr,nodes[i]->u_rank,d-dinc,&descRSr,Sr_upd,dinc,descRSr_upd,&tmpinfo);
        delete[] Sr_upd;
        updateColumns(&Sc,nodes[i]->v_rank,d-dinc,&descRSc,Sc_upd,dinc,descRSc_upd,&tmpinfo);
        delete[] Sc_upd;
        updateColumns(&Rr,nodes[i]->v_rank,d-dinc,&descRSc,Rr_upd,dinc,descRSc_upd,&tmpinfo);
        delete[] Rr_upd;
        updateColumns(&Rc,nodes[i]->u_rank,d-dinc,&descRSr,Rc_upd,dinc,descRSr_upd,&tmpinfo);
        delete[] Rc_upd;
      }

      /* Mark node as compressed */
      nodes[i]->numrand=d;
      nodes[i]->state=COMPRESSED;

      /* Processes out of the grid for the current might be active at other nodes.
       * They need to know the state, and ranks and indices if COMPRESSED.
       * They receive the info from proc_b.
       */
      if(myid==proc_b) {
        for(k=proc_g+1;k<=proc_e;k++)
          MPI_Send((void*) &nodes[i]->state,IONE,MPI_INTEGER,k,1,comm);

        if(nodes[i]->state==COMPRESSED) {
          for(k=proc_g+1;k<=proc_e;k++)
            MPI_Send((void*) &nodes[i]->u_rank,IONE,MPI_INTEGER,k,1,comm);
          for(k=proc_g+1;k<=proc_e;k++)
            MPI_Send((void*) &nodes[i]->u_rows,IONE,MPI_INTEGER,k,1,comm);
          if(nodes[i]->u_rank>0)
            for(k=proc_g+1;k<=proc_e;k++)
              MPI_Send((void*) nodes[i]->Ir,nodes[i]->u_rank,MPI_INTEGER,k,1,comm);

          for(k=proc_g+1;k<=proc_e;k++)
            MPI_Send((void*) &nodes[i]->v_rank,IONE,MPI_INTEGER,k,1,comm);
          for(k=proc_g+1;k<=proc_e;k++)
            MPI_Send((void*) &nodes[i]->v_rows,IONE,MPI_INTEGER,k,1,comm);
          if(nodes[i]->v_rank>0)
            for(k=proc_g+1;k<=proc_e;k++)
              MPI_Send((void*) nodes[i]->Ic,nodes[i]->v_rank,MPI_INTEGER,k,1,comm);
        }
      }
    } else {
      /* Processes out of the grid also participate to collective
       * calls to the user-given access routine.
       */
      if(dosync)
        participate_to_calls(elements,nb_calls_level[nodes[i]->level],comm);

      if(i==dim_tree-1) {
        nodes[i]->state=COMPRESSED;
        break;
      }

      /* Processes out of the grid for the current might be active at other nodes.
       * They need to know the state, and ranks and indices if COMPRESSED.
       * (otherwise they simply go to the parent node).
       * They receive the info from proc_b.
       */
      MPI_Recv((void*) &nodes[i]->state,IONE,MPI_INTEGER,proc_b,1,comm,&status);
      if(nodes[i]->state==COMPRESSED) {
        nodes[i]->numrand=d;
        MPI_Recv((void*) &nodes[i]->u_rank,IONE,MPI_INTEGER,proc_b,1,comm,&status);
        MPI_Recv((void*) &nodes[i]->u_rows,IONE,MPI_INTEGER,proc_b,1,comm,&status);
        if(nodes[i]->u_rank>0) {
          nodes[i]->Ir=new int[nodes[i]->u_rank];
          MPI_Recv((void*) nodes[i]->Ir,nodes[i]->u_rank,MPI_INTEGER,proc_b,1,comm,&status);
        }

        MPI_Recv((void*) &nodes[i]->v_rank,IONE,MPI_INTEGER,proc_b,1,comm,&status);
        MPI_Recv((void*) &nodes[i]->v_rows,IONE,MPI_INTEGER,proc_b,1,comm,&status);
        if(nodes[i]->v_rank>0) {
          nodes[i]->Ic=new int[nodes[i]->v_rank];
          MPI_Recv((void*) nodes[i]->Ic,nodes[i]->v_rank,MPI_INTEGER,proc_b,1,comm,&status);
        }
      }
    }
  }

  if(dim_tree!=1 && nodes[dim_tree-1]->state!=COMPRESSED) {
    /* Generate new random vectors and samples,
     * and distribute them conforming to the tree.
     */
    restart(&Rr_upd,&Sr_upd,&Rc_upd,&Sc_upd,&descRS_upd,n,dinc,A,descA,ginfo,nodes[dim_tree-1]->allctxt,&rgen,&tRand,&tGEMM);
    time_Rand+=tRand;
    time_GEMM+=tGEMM;
    /* Restart the tree */
    goto restart;
  }

  /* Reduce local maximum ranks */
  MPI_Allreduce(MPI_IN_PLACE,r,IONE,MPI_INTEGER,MPI_MAX,comm);

  *deff=d;

  if(elements==NULL) {
    delete[] A12;
    delete[] A21;
    if(subtree[2*myid]>=0)
      delete[] D; /* Keep D only for parallel leaves */
  }
  if(!RSprovided) {
    delete[] Rr;
    delete[] Rc;
    delete[] Sr;
    delete[] Sc;
  }
  if(dosync) {
    delete[] nb_calls_level;
  }

}

template<typename T, typename S> void HSS_tree<T,S>::participate_to_calls(void (*elements)(void *, int *, int *, T *, int *), int nbcalls, MPI_Comm comm) {
  /* When the element access routine is global, we need all the processes to
   * participate to *every call*. Therefore, at every level, a process has to
   * performs one (D) or two (B12, B21) calls to "elements" for itself but also
   * a bunch of calls to "help" the other processes, even though these calls
   * correspond to nodes that are not mapped on that process. This is what
   * this routine does: it performs nbcalls "dummy" calls to elements.
   *
   * The routine is a loop of nbcalls calls to elements. Before each call,
   * the processes receives the vectors I and J and the descriptor descB to
   * be passed to elements (the API is that when we call elements, all the
   * arguments are defined on all the processes, besides B=NULL for processes
   * out of the grid).
   */

  int i;
  int *I, *J;
  int descin[BLACSCTXTSIZE], descB[BLACSCTXTSIZE];
  int nI, nJ;
  int mb, nb;
  int source;
  MPI_Status status;
#if defined(debug)
  int myid;

  MPI_Comm_rank(comm,&myid);
#endif

  for(i=0;i<nbcalls;i++) {
    /* Main loop */
#if defined(debug)
    std::cout << myid << " waiting for information; call " << i+1 << "/" << nbcalls << std::endl;
#endif

    /* Wait for information */
    MPI_Recv((void *)descin,BLACSCTXTSIZE,MPI_INTEGER,MPI_ANY_SOURCE,2,comm,&status);
    source=status.MPI_SOURCE;
    nI=descin[BLACSm];
    nJ=descin[BLACSn];
    mb=descin[BLACSmb];
    nb=descin[BLACSnb];
    I=new int[nI];

#if defined(debug)
    std::cout << myid << " got nI = " << nI << ", nJ = " << nJ << " from " << source << std::endl;
#endif

    MPI_Recv((void *)I,nI,MPI_INTEGER,MPI_ANY_SOURCE,2,comm,&status);
    if(status.MPI_SOURCE!=source) {
      std::cout << "Fatal error in participate_to_calls" << std::endl;
      MPI_Abort(comm,2);
    }
    J=new int[nJ];
    MPI_Recv((void *)J,nJ,MPI_INTEGER,MPI_ANY_SOURCE,2,comm,&status);
    if(status.MPI_SOURCE!=source) {
      std::cout << "Fatal error in participate_to_calls" << std::endl;
      MPI_Abort(comm,2);
    }

    /* Participate to call */
#if defined(debug)
    std::cout << myid << " starting call" << std::endl;
#endif
    descset_(descB,&nI,&nJ,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
    elements(obj,I,J,NULL,descB);
    MPI_Barrier(comm);

    /* Cleanup */
#if defined(debug)
    std::cout << myid << " done with call" << std::endl;
#endif
    delete[] I;
    delete[] J;

  }


}

#endif // HSSR_PAR_COMPRESS_HPP
