/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef PAR_TOOLS_HPP
#define PAR_TOOLS_HPP

template<typename T> void mergeVector(T *X, int proc_b1, int proc_e1, int m1, int i1, int proc_b2, int proc_e2, int m2, int i2, int proc_bp, int proc_ep, int d, int *descX, T **Y, int **descY, int allctxt, gridinfo *ginfo, MPI_Comm comm) {
  /* Merging the rows of two m1 x d and m2 x d matrices ("vectors") into
   * a single (m1+m2) x d array. The opposite of splitVector.
   *
   * Input:
   *  - processes proc_b1...proc_e1 own the m1 x d piece (in X(i1:i1+m1-1,:)).
   *  - processes proc_b2...proc_e2 own the m2 x d piece (in X(i2:i2+m2-1,:)).
   *  - processes proc_bp...proc_ep own the result (in Y) at the end.
   *  - allctxt is a context that contains all the calling processes
   *  - gridinfo contains grid information for the grid associated with
   *  proc_bp...proc_ep.
   *  - Assumption: [proc_b1...proc_e1] U [proc_b2...proc_e2] included in
   *  [proc_bp...proc_ep].
   *
   * Output: processes proc_bp...proc_ep share a distributed version of
   * the (m1+m2) x d matrix (= [m1 x d; m2 x d]) in array Y.
   * NB: some of these processes might out of the grid and own nothing.
   *
   */

  int m;
  int mb, nb;
  int myrow, mycol, nprow, npcol;
  int idummy;
  int ctxt;
  int desc[BLACSCTXTSIZE];
  int locrY, loccY;
  int myid;
  int ierr;

  MPI_Comm_rank(comm,&myid);

  ctxt=ginfo->ctxt;
  nprow=ginfo->nprow;
  npcol=ginfo->npcol;
  myrow=ginfo->myrow;
  mycol=ginfo->mycol;
  mb=ginfo->mb;
  nb=ginfo->nb;

  /* Create space for the result */
  m=m1+m2;
  *descY=new int[BLACSCTXTSIZE];
  if(myid>=proc_bp && myid<=proc_ep) {
    locrY=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccY=numroc_(&d,&nb,&mycol,&IZERO,&npcol);
    *Y=new T[locrY*loccY];
    idummy=std::max(1,locrY);
    descinit_(*descY,&m,&d,&mb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
  } else {
    descset_(*descY,&m,&d,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
    *Y=NULL;
  }

  if(m1!=0) {
    /* First, redistribute the m1 x d piece; it goes in the first m1 rows of Y */
    if(myid>=proc_b1 && myid<=proc_e1) {
      copyDesc(descX,desc);
    } else {
      idummy=m1+i1-1;
      descset_(desc,&idummy,&d,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
    }
    pgemr2d(m1,d,X,i1,IONE,desc,*Y,IONE,IONE,*descY,allctxt);
  }

  if(m2!=0) {
    /* Second, redistribute the m2 x d piece; it goes in the last m2 rows of Y */
    if(myid>=proc_b2 && myid<=proc_e2) {
      copyDesc(descX,desc);
    } else {
      idummy=m2+i2-1;
      descset_(desc,&idummy,&d,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
    }
    pgemr2d(m2,d,X,i2,IONE,desc,*Y,m1+1,IONE,*descY,allctxt);
  }

}

template<typename T> void splitVector(T *X, int proc_b1, int proc_e1, int m1, int proc_b2, int proc_e2, int m2, int d, int *descX, T **Y, int **descY, int allctxt, gridinfo *ginfo1, gridinfo *ginfo2, MPI_Comm comm) {
  /* Splitting the rows an (m1+m2) x d matrix ("vector") into two m1 x d
   * and m2 x d arrays. The opposite of mergeVector.
   *
   * Input:
   *  - all the processes share a distributed version of the (m1+m2) x d matrix
   * (= [m1 x d; m2 x d]) in array X.
   *  - allctxt contains all the processes of the calling node.
   *  - gridinfo1 contains grid information for the grid associated with
   *  all the [proc_b1...proc_e1] processes.
   *  - gridinfo2 contains grid information for the grid associated with
   *  all the [proc_b2...proc_e2] processes.
   *
   * Output:
   *  - processes proc_b1...proc_e1 own the m1 x d piece (in Y).
   *  - processes proc_b2...proc_e2 own the m2 x d piece (in Y).
   *
   */

  int idummy;
  int desc[BLACSCTXTSIZE];
  int locrY, loccY;
  int myid;
  int ierr;

  MPI_Comm_rank(comm,&myid);

  *descY=new int[BLACSCTXTSIZE];

  /* First, redistribute the m1 x d piece; it goes in X for processes proc_b1...proc_e1 */
  if(myid>=proc_b1 && myid<=proc_e1) {
    locrY=numroc_(&m1,&ginfo1->mb,&ginfo1->myrow,&IZERO,&ginfo1->nprow);
    loccY=numroc_(&d,&ginfo1->nb,&ginfo1->mycol,&IZERO,&ginfo1->npcol);
    *Y=new T[locrY*loccY];
    idummy=std::max(1,locrY);
    descinit_(desc,&m1,&d,&ginfo1->mb,&ginfo1->nb,&IZERO,&IZERO,&ginfo1->ctxt,&idummy,&ierr);
    descinit_(*descY,&m1,&d,&ginfo1->mb,&ginfo1->nb,&IZERO,&IZERO,&ginfo1->ctxt,&idummy,&ierr);
  } else {
    descset_(desc,&m1,&d,&ginfo1->mb,&ginfo1->nb,&IZERO,&IZERO,&INONE,&IONE);
  }
  pgemr2d(m1,d,X,IONE,IONE,descX,*Y,IONE,IONE,desc,allctxt);

  /* Second, redistribute the m2 x d piece; it goes in X for processes proc_b2...proc_e2 */
  if(myid>=proc_b2 && myid<=proc_e2) {
    locrY=numroc_(&m2,&ginfo2->mb,&ginfo2->myrow,&IZERO,&ginfo2->nprow);
    loccY=numroc_(&d,&ginfo2->nb,&ginfo2->mycol,&IZERO,&ginfo2->npcol);
    *Y=new T[locrY*loccY];
    idummy=std::max(1,locrY);
    descinit_(desc,&m2,&d,&ginfo2->mb,&ginfo2->nb,&IZERO,&IZERO,&ginfo2->ctxt,&idummy,&ierr);
    descinit_(*descY,&m2,&d,&ginfo2->mb,&ginfo2->nb,&IZERO,&IZERO,&ginfo2->ctxt,&idummy,&ierr);
  } else {
    descset_(desc,&m2,&d,&ginfo2->mb,&ginfo2->nb,&IZERO,&IZERO,&INONE,&IONE);
  }
  pgemr2d(m2,d,X,m1+1,IONE,descX,*Y,IONE,IONE,desc,allctxt);

}

template<typename T, typename S> void parExtractSubmatrix(T* A, int m, int n, int *descA, int *Ir, int rr, int *Ic, int rc, T** B, gridinfo *ginfo) {
  /* Copy A[Ir,Ic] into B, with Ir and Ic sets of indices of 1:n of size rr and rc
   * rc respectively; if rr (resp. rc) =-1, extract all rows (resp. columns)
   *
   * Note that this routine implies communications. We cannot rely on placpy
   * that just performs independent local copies; instead we use pgeadd
   * we alpha=ONE and beta=ZERO (i.e. input submatrix erases output submatrix).
   *
   * NB: Ir and Ic contain 1-based values.
   * */
  int mb, nb;
  int locr, locc;
  int myrow, mycol, nprow, npcol;
  int idummy, ctxt, ierr, desc[BLACSCTXTSIZE];
  int i;
  T *tmp;

  T ZERO=0;
  T  ONE=1;

  ctxt=ginfo->ctxt;
  nprow=ginfo->nprow;
  npcol=ginfo->npcol;
  myrow=ginfo->myrow;
  mycol=ginfo->mycol;
  mb=ginfo->mb;
  nb=ginfo->nb;

  if(rr==0 || rc==0) {
    *B=NULL;
    return;
  }

  if(rr==-1 && rc==-1) {
    locr=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    locc=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    *B=new T[locr*locc];
    idummy=std::max(1,locr);
    descinit_(desc,&m,&n,&mb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
    pgeadd<T>('N',m,n,ONE,A,IONE,IONE,descA,ZERO,*B,IONE,IONE,desc);
  } else if (rr==-1) {
    locr=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    locc=numroc_(&rc,&nb,&mycol,&IZERO,&npcol);
    *B=new T[locr*locc];
    idummy=std::max(1,locr);
    descinit_(desc,&m,&rc,&mb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
    for(i=0;i<rc;i++)
      pgeadd<T>('N',m,IONE,ONE,A,IONE,Ic[i],descA,ZERO,*B,IONE,i+1,desc);
  } else if (rc==-1) {
    locr=numroc_(&rr,&mb,&myrow,&IZERO,&nprow);
    locc=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    *B=new T[locr*locc];
    idummy=std::max(1,locr);
    descinit_(desc,&rr,&n,&mb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
    for(i=0;i<rr;i++)
      pgeadd<T>('N',IONE,n,ONE,A,Ir[i],IONE,descA,ZERO,*B,i+1,IONE,desc);
  } else {
    /* Nothing very smart here, extract rows then columns */
    parExtractSubmatrix<T,S>(A,m,n,descA,Ir,rr,NULL,-1,&tmp,ginfo);
    locr=numroc_(&rr,&mb,&myrow,&IZERO,&nprow);
    locc=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    idummy=std::max(1,locr);
    descinit_(desc,&rr,&n,&mb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
    parExtractSubmatrix<T,S>(tmp,rr,n,desc,NULL,-1,Ic,rc,B,ginfo);
    delete[] tmp;
  }

}

inline void exchangeIndices(int **I1, int *r1, int proc_b1, int proc_e1, int **I2, int *r2, int proc_b2, int proc_e2, MPI_Comm comm) {
 /* Exchange routine.
  * Input:
  *  - processes proc_b1...proc_e1 own I1 (size r1).
  *  - processes proc_b2...proc_e2 own I2 (size r2).
  *
  * Output: all the processes own I1 and I2.
  *
  * Collective operations use a loop of sends with the same source. This could
  * (should?) be improved, at least by using different sources.
  *
  */

  int myid;
  int i;
  MPI_Status status;

  MPI_Comm_rank(comm,&myid);

  /* First, process proc_b1 send I1 to processes proc_b2...proc_e2.
   * Start with the size (r1), then the array.
   */
  if(myid==proc_b1) {
    for(i=proc_b2;i<=proc_e2;i++)
      MPI_Send((void*) r1,IONE,MPI_INTEGER,i,0,comm);
  }
  else if(myid>=proc_b2 && myid<=proc_e2){
      MPI_Recv((void*) r1,IONE,MPI_INTEGER,proc_b1,0,comm,&status);
  }
  if(*r1>0) {
    if(myid==proc_b1) {
      for(i=proc_b2;i<=proc_e2;i++)
        MPI_Send((void*) *I1,*r1,MPI_INTEGER,i,0,comm);
    }
    else if(myid>=proc_b2 && myid<=proc_e2){
        *I1=new int[*r1];
        MPI_Recv((void*) *I1,*r1,MPI_INTEGER,proc_b1,0,comm,&status);
    }
  }

  /* Second, process proc_b2 send I2 to processes proc_b1...proc_e1.
   * Start with the size (r2), then the array.
   */
  if(myid==proc_b2) {
    for(i=proc_b1;i<=proc_e1;i++)
      MPI_Send((void*) r2,IONE,MPI_INTEGER,i,0,comm);
  }
  else if(myid>=proc_b1 && myid<=proc_e1){
    MPI_Recv((void*) r2,IONE,MPI_INTEGER,proc_b2,0,comm,&status);
  }
  if(*r2>0) {
    if(myid==proc_b2) {
      for(i=proc_b1;i<=proc_e1;i++)
        MPI_Send((void*) *I2,*r2,MPI_INTEGER,i,0,comm);
    }
    else if(myid>=proc_b1 && myid<=proc_e1){
      *I2=new int[*r2];
      MPI_Recv((void*) *I2,*r2,MPI_INTEGER,proc_b2,0,comm,&status);
    }
  }

}

inline void exchangeInteger(int *x1, int proc_b1, int proc_e1, int *x2, int proc_b2, int proc_e2, MPI_Comm comm) {
 /* Exchange routine.
  * Input:
  *  - processes proc_b1...proc_e1 own x1.
  *  - processes proc_b2...proc_e2 own x2.
  *
  * Output: all the processes own x1 and x2.
  *
  * Collective operations use a loop of sends with the same source. This could
  * (should?) be improved, at least by using different sources.
  *
  */

  int myid;
  int i;
  MPI_Status status;

  MPI_Comm_rank(comm,&myid);

  /* First, process proc_b1 send x1 to processes proc_b2...proc_e2. */
  if(myid==proc_b1) {
    for(i=proc_b2;i<=proc_e2;i++)
      MPI_Send((void*) x1,IONE,MPI_INTEGER,i,0,comm);
  }
  else if(myid>=proc_b2 && myid<=proc_e2){
      MPI_Recv((void*) x1,IONE,MPI_INTEGER,proc_b1,0,comm,&status);
  }

  /* Second, process proc_b2 send x2 to processes proc_b1...proc_e1. */
  if(myid==proc_b2) {
    for(i=proc_b1;i<=proc_e1;i++)
      MPI_Send((void*) x2,IONE,MPI_INTEGER,i,0,comm);
  }
  else if(myid>=proc_b1 && myid<=proc_e1){
    MPI_Recv((void*) x2,IONE,MPI_INTEGER,proc_b2,0,comm,&status);
  }

}

template<typename T> inline void updateColumns(T **X, int n, int d, int **descX, T *Xupd, int dupd, int *descXupd, gridinfo *ginfo) {
  /* X <- [ X Xupd ] with X n x d in input and n x (d+dupd) in output, Xupd n x dupd. */
  T * Xnew;
  int *descXnew;
  int locr, locc, dnew, idummy;
  int ctxt, nprow, npcol, myrow, mycol, mb, nb;
  int ierr;

  T ZERO=0;
  T  ONE=1;

  ctxt=ginfo->ctxt;
  nprow=ginfo->nprow;
  npcol=ginfo->npcol;
  myrow=ginfo->myrow;
  mycol=ginfo->mycol;
  mb=ginfo->mb;
  nb=ginfo->nb;

  dnew=d+dupd;
  locr=numroc_(&n,&mb,&myrow,&IZERO,&nprow);
  locc=numroc_(&dnew,&nb,&mycol,&IZERO,&npcol);
  idummy=std::max(locr,1);
  descXnew=new int[BLACSCTXTSIZE];
  descinit_(descXnew,&n,&dnew,&mb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);

  Xnew=new T[locr*locc];
  pgeadd('N',n,d,ONE,*X,IONE,IONE,*descX,ZERO,Xnew,IONE,IONE,descXnew);
  if(dupd!=0)
    pgeadd('N',n,dupd,ONE,Xupd,IONE,IONE,descXupd,ZERO,Xnew,IONE,d+1,descXnew);

  *X=Xnew;
  *descX=descXnew;

}

template<typename T> inline void generateSamples(T **Rr, T **Sr, T** Rc, T **Sc, int **descRS, int n, int d, T* A, int *descA, gridinfo *ginfo, StrumpackRand *r, MPI_Comm comm, double *tRand, double *tGEMM) {
  int locrRS, loccRS, idummy;
  int ctxt, nprow, npcol, myrow, mycol, mb, nb;
  int myid;
  int ierr;
  double t;

  T ZERO=0;
  T  ONE=1;

  ctxt=ginfo->ctxt;
  nprow=ginfo->nprow;
  npcol=ginfo->npcol;
  myrow=ginfo->myrow;
  mycol=ginfo->mycol;
  mb=ginfo->mb;
  nb=ginfo->nb;
  *tRand=0;
  *tGEMM=0;
  MPI_Comm_rank(comm,&myid);

  /* Step 1: generating random vectors Rr and Rc */
  *descRS=new int[BLACSCTXTSIZE];
  if(myid<nprow*npcol) {
    locrRS=numroc_(&n,&mb,&myrow,&IZERO,&nprow);
    loccRS=numroc_(&d,&nb,&mycol,&IZERO,&npcol);
    idummy=std::max(1,locrRS);
    descinit_(*descRS,&n,&d,&mb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);

    *Rr=NULL;
    *Rr=new T[locrRS*loccRS];
    t=MPI_Wtime();
    parDenseRand(*Rr,n,d,mb,nb,myrow,mycol,nprow,npcol,r);
    *tRand+=MPI_Wtime()-t;

    *Rc=NULL;
    *Rc=new T[locrRS*loccRS];
    t=MPI_Wtime();
    parDenseRand(*Rc,n,d,mb,nb,myrow,mycol,nprow,npcol,r);
    *tRand+=MPI_Wtime()-t;

#if defined(debug)
    printMatrix("Rr: "+SSTR(myid),*Rr,locrRS,loccRS,locrRS);
    printMatrix("Rc: "+SSTR(myid),*Rc,locrRS,loccRS,locrRS);
#endif
  } else {
    *Rr=NULL;
    *Rc=NULL;
    descset_(*descRS,&n,&d,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
  }

  /* Step 2: generating samples: Sr=A*Rr, Sc=A'*Rc */
  if(myid<nprow*npcol) {
    *Sr=NULL;
    *Sr=new T[locrRS*loccRS];
    t=MPI_Wtime();
    pgemm<T>('N','N',n,d,n,ONE,A,IONE,IONE,descA,*Rr,IONE,IONE,*descRS,ZERO,*Sr,IONE,IONE,*descRS);
    *tGEMM+=MPI_Wtime()-t;

    *Sc=NULL;
    *Sc=new T[locrRS*loccRS];
    t=MPI_Wtime();
    pgemm<T>('C','N',n,d,n,ONE,A,IONE,IONE,descA,*Rc,IONE,IONE,*descRS,ZERO,*Sc,IONE,IONE,*descRS);
    *tGEMM+=MPI_Wtime()-t;

#if defined(debug)
    printMatrix("Sr: "+SSTR(myid),*Sr,locrRS,loccRS,locrRS);
    printMatrix("Sc: "+SSTR(myid),*Sc,locrRS,loccRS,locrRS);
#endif
  } else {
    *Sr=NULL;
    *Sc=NULL;
  }
}

#endif // PAR_TOOLS_HPP
