/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef TOOLS_HPP
#define TOOLS_HPP

#define SSTR(x) dynamic_cast<std::ostringstream&>(std::ostringstream() << std::dec << x).str()

typedef struct gridinfo{
  int ctxt;
  int nprow;
  int npcol;
  int myrow;
  int mycol;
  int mb;
  int nb;
} gridinfo;

inline void copyDesc(int *descX, int *descY) {
  int i;

  for(i=0;i<BLACSCTXTSIZE;i++)
    descY[i]=descX[i];

}

template<typename T> void copyTrans(int m, int n, T *A, int lda, T *B, int ldb) {
  /* Transpose and copy A (m x n) into B=A' (n x m) */
  int i, j;

  for(j=0;j<m;j++) {
    for(i=0;i<n;i++)
      B[j*ldb+i]=A[i*lda+j];
    lacgv(n,&B[j*ldb],1);
  }
}

template<typename T> void extractSubmatrix(T *A, int m, int n, int lda, int *Ir, int rr, int *Ic, int rc, T **B) {
  /* Copy A[Ir,Ic] into B, with Ir and Ic sets of indices of 1:n of size rr and rc
   * rc respectively; if rr (resp. rc) =-1, extract all rows (resp. columns)
   *
   * NB: Ir and Ic contain 1-based values.
   *
   */
  int i,j;

  if(rr==-1 && rc==-1) {
    *B=new T[m*n];
    lacpy('N',m,n,A,lda,*B,m);
  } else if(rr==-1) {
    *B=new T[m*rc];
    for(i=0;i<m;i++)
      for(j=0;j<rc;j++)
        (*B)[j*m+i]=A[(Ic[j]-1)*lda+i];
  } else if(rc==-1) {
    *B=new T[rr*n];
    for(i=0;i<rr;i++)
      for(j=0;j<n;j++)
        (*B)[j*rr+i]=A[j*lda+(Ir[i]-1)];
  } else {
    *B=new T[rr*rc];
    for(i=0;i<rr;i++)
      for(j=0;j<rc;j++)
        (*B)[j*rr+i]=A[(Ic[j]-1)*lda+(Ir[i]-1)];
  }
}

template<typename T> void printMatrix(std::string str, T *A, int m, int n, int lda) {
  int i,j;
  if(m*n<=0)
    return;
  if(str.length()>0)
  std::cout << str << ":" << std::endl;
  for(i=0;i<m;i++) {
    for(j=0;j<n;j++) {
      std::cout << std::right << std::setw(12) << A[j*lda+i] << " ";
    }
    std::cout << std::endl;
  }
  std::cout << std::endl;

}

template<typename T> inline void denseRand(T*, int, int, StrumpackRand *r);
template<> inline void denseRand<double>(double *R, int m, int n, StrumpackRand *r) {
  int i;
  for(i=0;i<m*n;i++) {
#if defined(debugmatrix)
    R[i]=static_cast<double>(i+1);
#else
    R[i]=(*r)();
#endif
  }
}
template<> inline void denseRand<float>(float *R, int m, int n, StrumpackRand *r) {
  int i;
  for(i=0;i<m*n;i++) {
#if defined(debugmatrix)
    R[i]=static_cast<float>(i+1);
#else
    R[i]=static_cast<float>((*r)());
#endif
  }
}
template<> inline void denseRand<dcomplex>(dcomplex *R, int m, int n, StrumpackRand *r) {
  int i;
  for(i=0;i<m*n;i++) {
#if defined(debugmatrix)
    R[i]=dcomplex(i+1,i);
#else
    R[i]=dcomplex((*r)(),(*r)());
#endif
  }
}
template<> inline void denseRand<scomplex>(scomplex *R, int m, int n, StrumpackRand *r) {
  int i;
  for(i=0;i<m*n;i++) {
#if defined(debugmatrix)
    R[i]=scomplex(i+1,i);
#else
    R[i]=scomplex((*r)(),(*r)());
#endif
  }
}

template<typename T> inline void parDenseRand(T*, int, int, int, int, int, int, int, int, StrumpackRand *reed);
template<> inline void parDenseRand<  double>(  double *R, int m, int n, int mb, int nb, int myrow, int mycol, int nprow, int npcol, StrumpackRand *r) {
  int i, locrR, loccR;
#if defined(debugmatrix)
  int j;
#endif
  locrR=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
  loccR=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
#if defined(debugmatrix)
  for(i=1;i<=m;i++)
    for(j=1;j<=n;j++)
      if(myrow==indxg2p_(&i,&mb,&myrow,&IZERO,&nprow) && mycol==indxg2p_(&j,&nb,&mycol,&IZERO,&npcol))
        R[(indxg2l_(&j,&nb,&mycol,&IZERO,&npcol)-1)*locrR+indxg2l_(&i,&mb,&myrow,&IZERO,&nprow)-1]=(j-1)*m+i;
#else
  for(i=0;i<locrR*loccR;i++)
    R[i]=static_cast<double>((*r)());
#endif
}
template<> inline void parDenseRand<   float>(   float *R, int m, int n, int mb, int nb, int myrow, int mycol, int nprow, int npcol, StrumpackRand *r) {
  int i, locrR, loccR;
#if defined(debugmatrix)
  int j;
#endif
  locrR=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
  loccR=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
#if defined(debugmatrix)
  for(i=1;i<=m;i++)
    for(j=1;j<=n;j++)
      if(myrow==indxg2p_(&i,&mb,&myrow,&IZERO,&nprow) && mycol==indxg2p_(&j,&nb,&mycol,&IZERO,&npcol))
        R[(indxg2l_(&j,&mb,&mycol,&IZERO,&npcol)-1)*locrR+indxg2l_(&i,&nb,&myrow,&IZERO,&nprow)-1]=(j-1)*m+i;
#else
  for(i=0;i<locrR*loccR;i++)
    R[i]=static_cast<float>((*r)());
#endif
}
template<> inline void parDenseRand<dcomplex>(dcomplex *R, int m, int n, int mb, int nb, int myrow, int mycol, int nprow, int npcol, StrumpackRand *r) {
  int i, locrR, loccR;
#if defined(debugmatrix)
  int j;
#endif
  locrR=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
  loccR=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
#if defined(debugmatrix)
  for(i=1;i<=m;i++)
    for(j=1;j<=n;j++)
      if(myrow==indxg2p_(&i,&mb,&myrow,&IZERO,&nprow) && mycol==indxg2p_(&j,&nb,&mycol,&IZERO,&npcol)) {
        R[(indxg2l_(&j,&mb,&mycol,&IZERO,&npcol)-1)*locrR+indxg2l_(&i,&nb,&myrow,&IZERO,&nprow)-1]=dcomplex((j-1)*m+i,(j-1)*m+i-1);
      }
#else
  for(i=0;i<locrR*loccR;i++) {
    R[i]=dcomplex((*r)(),(*r)());
  }
#endif
}
template<> inline void parDenseRand<scomplex>(scomplex *R, int m, int n, int mb, int nb, int myrow, int mycol, int nprow, int npcol, StrumpackRand *r) {
  int i, locrR, loccR;
#if defined(debugmatrix)
  int j;
#endif
  locrR=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
  loccR=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
#if defined(debugmatrix)
  for(i=1;i<=m;i++)
    for(j=1;j<=n;j++)
      if(myrow==indxg2p_(&i,&mb,&myrow,&IZERO,&nprow) && mycol==indxg2p_(&j,&nb,&mycol,&IZERO,&npcol)) {
        R[(indxg2l_(&j,&mb,&mycol,&IZERO,&npcol)-1)*locrR+indxg2l_(&i,&nb,&myrow,&IZERO,&nprow)-1]=scomplex((j-1)*m+i,(j-1)*m+i-1);
      }
#else
  for(i=0;i<locrR*loccR;i++) {
    R[i]=scomplex((*r)(),(*r)());
  }
#endif
}

#endif // TOOLS_HPP
