/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef GEOMETRIC_REORDERING_HPP
#define GEOMETRIC_REORDERING_HPP

#include "MatrixReordering.hpp"
#include <array>

namespace strumpack {

  template<typename scalar_t,typename real_t,typename integer_t> class GeometricReordering
    : public MatrixReordering<scalar_t,real_t,integer_t> {
  public:
    GeometricReordering();
    ~GeometricReordering();
    std::string get_name() { return std::string("Geometric"); };

    virtual int nested_dissection(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool diag,
				  int stratnum, int stratpar, bool verbose) {
      std::cout << "Error: Geometric reordering does not support general sparse matrices." << std::endl
		<< "       Specify the grid dimensions or use Metis or Scotch." << std::endl;
      exit(1);
      return 1;
    }
    virtual int nested_dissection(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool diag,
				  int stratnum, int stratpar, bool verbose,
				  int nx, int ny, int nz);
  protected:
    void recursive_nested_dissection(std::array<integer_t,3> n0,
				     std::array<integer_t,3> dims,
				     std::array<integer_t,3> ld,
				     integer_t& order_begin,
				     int stratpar, int depth=0);
  };

  template<typename scalar_t,typename real_t,typename integer_t>
  GeometricReordering<scalar_t,real_t,integer_t>::GeometricReordering() {}

  template<typename scalar_t,typename real_t,typename integer_t>
  GeometricReordering<scalar_t,real_t,integer_t>::~GeometricReordering() {}

  // produce the nested dissection ordering and separator information
  // for model problem NX x NY x NZ grid with a stencil which is 3 points per dimension
  template<typename scalar_t,typename real_t,typename integer_t> int
  GeometricReordering<scalar_t,real_t,integer_t>::nested_dissection
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool diag,
   int stratnum, int stratpar, bool verbose, int nx, int ny, int nz) {
    integer_t n = A->size();
    assert(A->size() == n);
    this->order.resize(n);
    this->iorder.resize(n);
    this->sep_ptr.push_back(0);
    this->nbsep = 0;
    integer_t order_begin = 0;
    recursive_nested_dissection({{0, 0, 0}}, {{nx, ny, nz}}, {{nx, ny, nz}},
				order_begin, stratpar);
    this->parent.shrink_to_fit();
    this->sep_ptr.shrink_to_fit();
    this->left_child.shrink_to_fit();
    this->right_child.shrink_to_fit();
    this->level.shrink_to_fit();
    integer_t max_level = this->level[0];
    std::for_each(this->level.begin(), this->level.end(), [&](integer_t& l){ l=max_level-l+1; });
    this->nested_dissection_print(n, 7*nx*ny*nz - 2*nx*ny - 2*nx*nz - 2*ny*nz, 0, 0, verbose);
    return 0;
  }

  /**
   *  compute nested-dissection reordering for a (dims[0] x dims[1] x dims[2]) mesh
   *   the mesh is a submesh of a regular (ld[0] x ld[1] x ld[2]) mesh,
   *   starting at corner {n0[0], n0[1], n0[2]}
   *  the permutation is computed in this->order, inverse permutation in this->iorder
   *  the parent info in this->parent
   *  the separator sizes in this->sep_ptr
   *  the child info in this->left_child / this_right_child
   */
  template<typename scalar_t,typename real_t,typename integer_t> void
  GeometricReordering<scalar_t,real_t,integer_t>::recursive_nested_dissection
  (std::array<integer_t,3> n0, std::array<integer_t,3> dims, std::array<integer_t,3> ld,
   integer_t& order_begin, int stratpar, int depth) {
    integer_t N = dims[0]*dims[1]*dims[2];
    // d: dimension along which to split
    int d = std::distance(dims.begin(), std::max_element(dims.begin(), dims.end()));

    if (dims[d] < 3 || N <= stratpar) {
      for (integer_t z=n0[2]; z<n0[2]+dims[2]; z++)
	for (integer_t y=n0[1]; y<n0[1]+dims[1]; y++)
	  for (integer_t x=n0[0]; x<n0[0]+dims[0]; x++) {
	    integer_t ind = x + y*ld[0] + z*ld[0]*ld[1];
	    this->order[ind] = order_begin;
	    this->iorder[order_begin] = ind;
	    order_begin++;
	  }
      this->sep_ptr.push_back(this->sep_ptr.back() + N);
      this->level.push_back(depth);
      this->left_child.push_back(-1);
      this->right_child.push_back(-1);
      this->parent.push_back(-1);
      this->nbsep++;
    } else {
      // part 1
      std::array<integer_t,3> part_begin(n0);
      std::array<integer_t,3> part_size(dims);
      part_size[d] = dims[d]/2;
      recursive_nested_dissection(part_begin, part_size, ld, order_begin, stratpar, depth+1);
      integer_t left_root_id = this->nbsep - 1;

      // part 2
      part_begin[d] = n0[d] + dims[d]/2 + 1;
      part_size[d] = dims[d] - 1 - dims[d]/2;
      recursive_nested_dissection(part_begin, part_size, ld, order_begin, stratpar, depth+1);
      this->parent[left_root_id] = this->nbsep;
      this->parent[this->nbsep-1] = this->nbsep;

      // separator
      part_begin[d] = n0[d] + dims[d]/2;
      part_size[d] = 1;
      for (integer_t z=part_begin[2]; z<part_begin[2]+part_size[2]; z++)
	for (integer_t y=part_begin[1]; y<part_begin[1]+part_size[1]; y++)
	  for (integer_t x=part_begin[0]; x<part_begin[0]+part_size[0]; x++) {
	    integer_t ind = x + y*ld[0] + z*ld[0]*ld[1];
	    this->order[ind] = order_begin;
	    this->iorder[order_begin] = ind;
	    order_begin++;
	  }
      this->sep_ptr.push_back(this->sep_ptr.back() + N/dims[d]);
      this->left_child.push_back(left_root_id);
      this->right_child.push_back(this->nbsep-1);
      this->parent.push_back(-1);
      this->level.push_back(depth);
      this->nbsep++;
    }
  }

} // end namespace strumpack

#endif
