/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef GEOMETRIC_REORDERING_MPI_HPP
#define GEOMETRIC_REORDERING_MPI_HPP

#include "MatrixReordering.hpp"
#include "MPI_wrapper.hpp"
#include <array>

namespace strumpack {

  template<typename scalar_t,typename real_t,typename integer_t> class GeometricReorderingMPI
    : public MatrixReorderingMPI<scalar_t,real_t,integer_t> {
  public:
    GeometricReorderingMPI(MPI_Comm c) : MatrixReorderingMPI<scalar_t,real_t,integer_t>(c) {}
    ~GeometricReorderingMPI() {}
    std::string get_name() { return std::string("GeometricMPI"); };

    virtual int nested_dissection(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool diag,
				  int stratnum, int stratpar, bool verbose) {
      std::cout << "Error: Geometric reordering does not support general sparse matrices." << std::endl
		<< "       Specify the grid dimensions or use Metis or Scotch." << std::endl;
      exit(1);
      return 1;
    }
    virtual int nested_dissection(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool diag,
				  int stratnum, int stratpar, bool verbose,
				  int nx, int ny, int nz);
  protected:
    void recursive_nested_dissection(integer_t lo, integer_t hi,
				     std::array<integer_t,3> n0,
				     std::array<integer_t,3> dims,
				     std::array<integer_t,3> ld,
				     integer_t& order_begin,
				     int stratpar, int depth,
				     integer_t dsep_id, integer_t& dsep_leaf_id,
				     int P, int rank);
  };

  // produce the nested dissection ordering and separator information
  // for model problem NX x NY x NZ grid with a stencil which is 3 points per dimension
  template<typename scalar_t,typename real_t,typename integer_t> int
  GeometricReorderingMPI<scalar_t,real_t,integer_t>::nested_dissection
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool diag,
   int stratnum, int stratpar, bool verbose, int nx, int ny, int nz) {
    assert(A->size() == nx * ny * nz);
    CSRMatrixMPI<scalar_t,real_t,integer_t>* A_mpi = dynamic_cast<CSRMatrixMPI<scalar_t,real_t,integer_t>*>(A);
    assert(A_mpi != NULL);

    auto P = mpi_nprocs(this->comm);
    auto rank = mpi_rank(this->comm);
    auto N = A_mpi->size();
    this->order.resize(N);
    this->iorder.resize(N);
    this->sep_ptr.push_back(0);
    this->dist_sep_ptr.push_back(0);
    this->nbsep = 0;
    this->dist_nr_sep = 0;
    integer_t order_begin = 0, dsep_leaf_id = 0;
    recursive_nested_dissection(A_mpi->begin_row(), A_mpi->end_row(),
				{{0, 0, 0}}, {{nx, ny, nz}}, {{nx, ny, nz}},
				order_begin, stratpar, 0, 1, dsep_leaf_id, P, rank);
    this->get_local_graphs(A_mpi);
    this->nested_dissection_print(A->size(), 7*nx*ny*nz - 2*nx*ny - 2*nx*nz - 2*ny*nz,
				  stratnum, stratpar, verbose);
    return 0;
  }


  /**
   * The distributed tree should have P leafs.  Lets number the
   * distributed separators level by level, root=0.  For instance: for
   * P=5, the distributed tree will look like:
   *                   1
   *                  / \
   *                 /   \
   *                2     3
   *               / \   / \
   *              4   5 6   7
   *             / \
   *            8   9
   *
   * The number of the node is given by dist_sep_id, left child will has
   * id 2*dist_sep_id, right child 2*dist_sep_id+1.  Nodes for which
   * P<=dist_sep_id<2*P, are leafs of the distributed separator tree,
   * and they form the roots of the local subtree for process
   * dist_sep_id-P.
   *
   * Furthermore, each proces has a local tree rooted at one of the
   * leafs of the distributed tree.  To have the same convention as used
   * for PTScotch, the leafs are assigned to procs as iin a postordering
   * of the distributed tree, hence for this example, proc_dist_sep:
   *                   3
   *                  / \
   *                 /   \
   *                1     2
   *               / \   / \
   *              0   2 3   4
   *             / \
   *            0   1
   */
  template<typename scalar_t,typename real_t,typename integer_t> void
  GeometricReorderingMPI<scalar_t,real_t,integer_t>::recursive_nested_dissection
  (integer_t lo, integer_t hi, std::array<integer_t,3> n0, std::array<integer_t,3> dims,
   std::array<integer_t,3> ld, integer_t& order_begin, int stratpar, int depth,
   integer_t dsep_id, integer_t& dsep_leaf_id, int P, int rank) {
    integer_t N = dims[0]*dims[1]*dims[2];
    // d: dimension along which to split
    int d = std::distance(dims.begin(), std::max_element(dims.begin(), dims.end()));

    bool dsep = dsep_id < 2*P;
    bool dsep_leaf = dsep && dsep_id >= P;
    bool local_tree = dsep_id >= P && dsep_leaf_id == rank;

    if (dims[d] < 3 || N <= stratpar) {
      for (integer_t z=n0[2]; z<n0[2]+dims[2]; z++)
	for (integer_t y=n0[1]; y<n0[1]+dims[1]; y++)
	  for (integer_t x=n0[0]; x<n0[0]+dims[0]; x++) {
	    integer_t ind = x + y*ld[0] + z*ld[0]*ld[1];
	    this->order[ind] = order_begin;
	    this->iorder[order_begin] = ind;
	    order_begin++;
	  }

      if (dsep) {
	this->dist_sep_ptr.push_back(this->dist_sep_ptr.back() + N);
	this->dist_level.push_back(depth);
	this->dist_lchild.push_back(-1);
	this->dist_rchild.push_back(-1);
	this->dist_parent.push_back(-1);
	this->dist_nr_sep++;
      }
      if (local_tree) {
	this->sep_ptr.push_back(this->sep_ptr.back() + N);
	this->level.push_back(depth);
	this->left_child.push_back(-1);
	this->right_child.push_back(-1);
	this->parent.push_back(-1);
	this->nbsep++;
      }
    } else {
      // part 1/left
      std::array<integer_t,3> part_begin(n0);
      std::array<integer_t,3> part_size(dims);
      part_size[d] = dims[d]/2;
      recursive_nested_dissection(lo, hi, part_begin, part_size, ld, order_begin,
				  stratpar, depth+1, 2*dsep_id, dsep_leaf_id, P, rank);
      integer_t left_root_id_dist = this->dist_nr_sep - 1;
      integer_t left_root_id_local = this->nbsep - 1;

      // part 2/right
      part_begin[d] = n0[d] + dims[d]/2 + 1;
      part_size[d] = dims[d] - 1 - dims[d]/2;
      recursive_nested_dissection(lo, hi, part_begin, part_size, ld, order_begin,
				  stratpar, depth+1, 2*dsep_id+1, dsep_leaf_id, P, rank);
      if (dsep && !dsep_leaf) {
	this->dist_parent[left_root_id_dist] = this->dist_nr_sep;
	this->dist_parent[this->dist_nr_sep-1] = this->dist_nr_sep;
      }
      if (local_tree) {
	this->parent[left_root_id_local] = this->nbsep;
	this->parent[this->nbsep-1] = this->nbsep;
      }

      // separator
      part_begin[d] = n0[d] + dims[d]/2;
      part_size[d] = 1;
      for (integer_t z=part_begin[2]; z<part_begin[2]+part_size[2]; z++)
	for (integer_t y=part_begin[1]; y<part_begin[1]+part_size[1]; y++)
	  for (integer_t x=part_begin[0]; x<part_begin[0]+part_size[0]; x++) {
	    integer_t ind = x + y*ld[0] + z*ld[0]*ld[1];
	    this->order[ind] = order_begin;
	    this->iorder[order_begin] = ind;
	    order_begin++;
	  }

      if (dsep) {
	if (dsep_leaf) {
	  this->dist_sep_ptr.push_back(this->dist_sep_ptr.back() + N);
	  this->dist_lchild.push_back(-1);
	  this->dist_rchild.push_back(-1);
	} else {
	  this->dist_sep_ptr.push_back(this->dist_sep_ptr.back() + N/dims[d]);
	  this->dist_lchild.push_back(left_root_id_dist);
	  this->dist_rchild.push_back(this->dist_nr_sep-1);
	}
	this->dist_parent.push_back(-1);
	this->dist_level.push_back(depth);
	this->dist_nr_sep++;
      }
      if (local_tree) {
	this->sep_ptr.push_back(this->sep_ptr.back() + N/dims[d]);
	this->left_child.push_back(left_root_id_local);
	this->right_child.push_back(this->nbsep-1);
	this->parent.push_back(-1);
	this->level.push_back(depth);
	this->nbsep++;
      }
    }
    if (dsep && dsep_leaf) dsep_leaf_id++;
  }

} // end namespace strumpack

#endif
