/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef HSS_MATRIX_ULV_HPP
#define HSS_MATRIX_ULV_HPP

namespace strumpack {

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::ULV_factorization(int depth, bool isroot) {
    int info = 0;
    bool tasked = depth<params::task_recursion_cutoff_level;
    reallocate_workmemory(0, std::max(u_rows*v_rank, u_rank*u_rows));

    if (child.size() > 0) {
      if (tasked) {
	for (auto ch : child)
#pragma omp task firstprivate(ch) default(shared) final(depth >= params::task_recursion_cutoff_level-1) mergeable
	  ch->ULV_factorization(depth+1, false);
#pragma omp taskwait
      } else {
	for (auto ch : child) ch->ULV_factorization(depth, false);
      }
      D = new scalar_t[u_rows*u_rows];
      if (u_rows) {
	integer_t bpos = 0, ipos = 0;
	for (auto ci : child)
	  if (ci->u_rank) {
	    integer_t jpos = 0;
	    for (auto cj : child) {
	      if (ci == cj) blas::lacpy('A', ci->u_rank, ci->u_rank, ci->D, ci->u_rank, &D[ipos*u_rows+jpos], u_rows);
	      else if (cj->u_rank>0 && ci->v_rank>0) {
		if (tasked) gemm_omp_task('N', 'C', cj->u_rank, ci->u_rank, ci->v_rank, scalar_t(1.), &B[bpos], cj->u_rank,
					  &(ci->Vt[ci->u_rows-ci->u_rank]), ci->u_rows, scalar_t(0.), &D[ipos*u_rows+jpos], u_rows, depth);
		else blas::gemm('N', 'C', cj->u_rank, ci->u_rank, ci->v_rank, scalar_t(1.), &B[bpos], cj->u_rank,
				&(ci->Vt[ci->u_rows-ci->u_rank]), ci->u_rows, scalar_t(0.), &D[ipos*u_rows+jpos], u_rows);
		bpos += cj->u_rank * ci->v_rank;
	      } jpos += cj->u_rank;
	    } ipos += ci->u_rank;
	  }
      }
      if (v_rank) { // put \hat{V} in work_v
	apply_V_permutation_forward(v_rank, V, v_rows);
	mat_zero(u_rows, v_rank, work_v, u_rows);
	integer_t vpos = 0, upos = 0;
	for (auto ch : child) {
	  if (ch->u_rank && ch->v_rank) {
	    if (tasked) gemm_omp_task('N', 'N', ch->u_rank, v_rank, ch->v_rank, scalar_t(1.), &(ch->Vt[ch->u_rows-ch->u_rank]),
				      ch->u_rows, &V[vpos], v_rows, scalar_t(0.), &work_v[upos], u_rows, depth);
	    else blas::gemm('N', 'N', ch->u_rank, v_rank, ch->v_rank, scalar_t(1.), &(ch->Vt[ch->u_rows-ch->u_rank]),
			    ch->u_rows, &V[vpos], v_rows, scalar_t(0.), &work_v[upos], u_rows);
	  }
	  upos += ch->u_rank; vpos += ch->v_rank;
	}
	apply_V_permutation_backward(v_rank, V, v_rows);
      }
    } else {
      if (v_rank) { // put \hat{V} in work_v
	blas::lacpy('A', u_rows, v_rank, V, u_rows, work_v, u_rows);
	apply_V_permutation_forward(v_rank, work_v, v_rows);
      }
    }

    Vt = new scalar_t[u_rows*v_rank];
    if (isroot) { // LU decomposition of D
      if (u_rows) {
	if (locr==NULL) locr = new int[u_rows];
	if (tasked) getrf_omp_task(u_rows, u_rows, D, u_rows, locr, &info, depth);
	else blas::getrf(u_rows, u_rows, D, u_rows, locr, &info);
	if (info){ std::cerr << "In HSSMatrix::ULV_factorization: LU factorization failed with info=" << info << std::endl; exit(1); }
	blas::lacpy('A', u_rows, v_rank, work_v, u_rows, Vt, u_rows);
      }
    } else {
      apply_U_permutation_backward(u_rows, D, u_rows);
      if (u_rows>u_rank) {
	Q = new scalar_t[u_rows*u_rows];
	if (u_rank) {
	  if (tasked) gemm_omp_task('N', 'N', u_rows-u_rank, u_rows, u_rank, scalar_t(-1.),
				    &U[u_rank], u_rows, D, u_rows, scalar_t(0.), Q, u_rows, depth);
	  else blas::gemm('N', 'N', u_rows-u_rank, u_rows, u_rank, scalar_t(-1.),
			  &U[u_rank], u_rows, D, u_rows, scalar_t(0.), Q, u_rows);
	  mat_add(u_rows-u_rank, u_rows, &D[u_rank], u_rows, scalar_t(1.), Q, u_rows);
	} else blas::lacpy('A', u_rows, u_rows, D, u_rows, Q, u_rows);
	scalar_t* tau = new scalar_t[std::max(integer_t(1), u_rows-u_rank)];
	// LQ(W_1) = [L 0] [ Q_1 ]
	//                 [ Q_2 ]
	if (tasked) blas::gelqfmod(u_rows-u_rank, u_rows, Q, u_rows, tau, &info, depth);
	else blas::gelqf(u_rows-u_rank, u_rows, Q, u_rows, tau, &info);
	if (info) { std::cerr << "In HSSMatrix::ULV_factorization: LQ factorization failed with info=" << info << std::endl; }
	L = new scalar_t[u_rows*u_rows];  // copy lower triangular to L
	blas::lacpy('A', u_rows-u_rank, u_rows-u_rank, Q, u_rows, L, u_rows);
	bool smalldiag = false;
	real_t sfmin = blas::lamch<real_t>('S');
	for (integer_t i=0; i<u_rows-u_rank; i++) if (std::abs(L[i+i*u_rows]) < sfmin) { smalldiag = true; break; }
	if (smalldiag) { std::cerr << "ERROR: small diagonal on L, from LQ" << std::endl; exit(1); }
	if (tasked) blas::xxglqmod(u_rows, u_rows, u_rows-u_rank, Q, u_rows, tau, &info, depth); // generate the full Q matrix in Q
	else blas::xxglq(u_rows, u_rows, u_rows-u_rank, Q, u_rows, tau, &info);
	if (info) { std::cerr << "In HSSMatrix::ULV_factorization: ORGLQ failed with info=" << info << std::endl; }
	delete[] tau;
	if (tasked) gemm_omp_task('N', 'N', u_rows, v_rank, u_rows, scalar_t(1.), Q, u_rows,  // \tilde{V} = Q \hat{V}  (\hat{V} sits in work_v)
				  work_v, u_rows, scalar_t(0.), Vt, u_rows, depth);
	else blas::gemm('N', 'N', u_rows, v_rank, u_rows, scalar_t(1.), Q, u_rows,  // \tilde{V} = Q \hat{V}  (\hat{V} sits in work_v)
			work_v, u_rows, scalar_t(0.), Vt, u_rows);
	if (u_rank) { // compute W_2 in work_v (D already permuted)
	  blas::lacpy('A', u_rank, u_rows, D, u_rows, work_v, u_rank);
	  if (tasked) gemm_omp_task('N', 'C', u_rank, u_rank, u_rows, scalar_t(1.), work_v, u_rank,  // D = W_2 Q_2^*
				    &(Q[u_rows-u_rank]), u_rows, scalar_t(0.), D, u_rank, depth);
	  else blas::gemm('N', 'C', u_rank, u_rank, u_rows, scalar_t(1.), work_v, u_rank,  // D = W_2 Q_2^*
			  &(Q[u_rows-u_rank]), u_rows, scalar_t(0.), D, u_rank);
	  if (tasked) gemm_omp_task('N', 'C', u_rank, u_rows-u_rank, u_rows, scalar_t(1.), work_v, u_rank,  // store W_2 Q_1^* under L
				    Q, u_rows, scalar_t(0.), &(L[u_rows-u_rank]), u_rows, depth);
	  else blas::gemm('N', 'C', u_rank, u_rows-u_rank, u_rows, scalar_t(1.), work_v, u_rank,  // store W_2 Q_1^* under L
			  Q, u_rows, scalar_t(0.), &(L[u_rows-u_rank]), u_rows);
	}
      } else // copy \hat{v} to \tilde{V}
	blas::lacpy('A', u_rows, v_rank, work_v, u_rows, Vt, u_rows);
    }
  }

  /**
   * TODO make thread-safe
   * if reduced_rhs != NULL, then also compute
   *   reduced_rhs = Vhat* y_0 + V* [z_0; z_1] for the root
   */
  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::ULV_forward_solve
  (scalar_t* f, scalar_t* reduced_rhs, int depth, bool isroot) {
    // [y; \tilde{f}_2] = work_u     should be u_rows
    // z = work_v                    should be v_rank
    reallocate_workmemory(u_rows, v_rank);
    bool tasked = depth<params::task_recursion_cutoff_level;

    if (child.size() > 0) {
      if (tasked) {
	for (auto ch : child)
#pragma omp task firstprivate(ch) default(shared) final(depth >= params::task_recursion_cutoff_level-1) mergeable
	  ch->ULV_forward_solve(f, NULL, depth+1, false);
#pragma omp taskwait
      } else {
	for (auto ch : child) ch->ULV_forward_solve(f, NULL, depth, false);
      }
      integer_t f_pos = 0;
      for (auto ci : child) {
	if (ci->u_rank>0) {
	  std::copy(ci->work_u+ci->u_rows-ci->u_rank, ci->work_u+ci->u_rows, work_u+f_pos);
	  if (ci->u_rows > ci->u_rank) // f -= W_i;2 Q*_i,1 y_i
	    gemv_omp_task('N', ci->u_rank, ci->u_rows-ci->u_rank, scalar_t(-1.),
			  ci->L+ci->u_rows-ci->u_rank, ci->u_rows, ci->work_u, 1,
			  scalar_t(1.), work_u+f_pos, 1, depth);
	  f_pos += ci->u_rank;
	}
      }
      integer_t B_pos = 0;
      for (auto ci : child) {
	f_pos = 0;
	for (auto cj : child) {
	  if (cj != ci) { // f -= B_{j,i} z_i
	    if (cj->u_rank && ci->v_rank)
	      gemv_omp_task('N', cj->u_rank, ci->v_rank, scalar_t(-1.), B+B_pos, cj->u_rank,
			    ci->work_v, 1, scalar_t(1.), work_u+f_pos, 1, depth);
	    B_pos += cj->u_rank * ci->v_rank;
	  } f_pos += cj->u_rank;
	}
      }
    } else std::copy(f+subset_begin, f+subset_begin+u_rows, work_u);

    if (isroot) { // PLU solve
      if (u_rows) {
	getrs_omp_task('N', u_rows, 1, D, u_rows, locr, work_u, u_rows, &flag, depth);
	if (reduced_rhs) {
	  // compute reduced_rhs = \hat{V}^* y_0 + V^* [z_0; z_1]
	  if (v_rank) {
	    gemv_omp_task('C', u_rows, v_rank, scalar_t(1.), Vt, u_rows,
			  work_u, 1, scalar_t(0.), reduced_rhs, 1, depth);
	    if (child.size() > 0) {
	      scalar_t* tmp = new scalar_t[v_rows];
	      std::copy(child[0]->work_v, child[0]->work_v + child[0]->v_rank, tmp);
	      std::copy(child[1]->work_v, child[1]->work_v + child[1]->v_rank, tmp + child[0]->v_rank);
	      apply_V(true, tmp, scalar_t(1.), reduced_rhs, depth);
	      delete[] tmp;
	    }
	  }
	}
      }
    } else {
      apply_U_permutation_backward(work_u);
      scalar_t* tmp = new scalar_t[std::max(u_rows,v_rows)]; // -> can use f for this??
      if (u_rows>u_rank) {
	if (u_rank) gemv_omp_task('N', u_rows-u_rank, u_rank, scalar_t(-1.),
				  U+u_rank, u_rows, work_u, 1, scalar_t(0.), tmp, 1, depth);
	else std::fill(tmp, tmp+u_rows-u_rank, scalar_t(0.));
	mat_add(u_rows-u_rank, integer_t(1), work_u+u_rank, u_rows, scalar_t(1.), tmp, u_rows);
	std::copy(work_u, work_u+u_rank, tmp+u_rows-u_rank);
	if (tasked) trsv_omp_task('L', 'N', 'N', u_rows-u_rank, L, u_rows, tmp, 1, depth);
	else blas::trsv('L', 'N', 'N', u_rows-u_rank, L, u_rows, tmp, 1);
	std::copy(tmp, tmp+u_rows, work_u);
      }
      if (v_rank) { // (z=work_v) = \tilde{V}*_{i;1} y_i
	if (u_rows>u_rank)
	  gemv_omp_task('C', u_rows-u_rank, v_rank, scalar_t(1.), Vt, u_rows,
			work_u, 1, scalar_t(0.), work_v, 1, depth);
	else std::fill(work_v, work_v+v_rank, scalar_t(0.));
	if (child.size() > 0) {
	  // (z=work_v) +=  V^* [ z_nu1; z_nu2; .. ]
	  integer_t pos = 0;
	  for (auto ch : child) { // tmp = [ work_v_nu1; work_v_nu2; .. ]
	    std::copy(ch->work_v, ch->work_v+ch->v_rank, tmp+pos);
	    pos += ch->v_rank;
	  }
	  apply_V(true, tmp, scalar_t(1.), work_v, depth);
	}
      }
      delete[] tmp;
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::ULV_backward_solve(scalar_t* f, int depth, bool isroot) {
    integer_t v_pos = 0;
    bool tasked = depth<params::task_recursion_cutoff_level;

    if (!child.empty()) {
      for (auto ch : child) {
	if (ch->u_rows-ch->u_rank > 0) { // otherwise full rank and Q=I
	  scalar_t* tmp = new scalar_t[ch->u_rows]; // use some available memory for this??
	  gemv_omp_task('C', ch->u_rows-ch->u_rank, ch->u_rows, scalar_t(1.), ch->Q, ch->u_rows,
			ch->work_u, 1, scalar_t(0.), tmp, 1, depth);
	  if (ch->u_rank)
	    gemv_omp_task('C', ch->u_rank, ch->u_rows, scalar_t(1.), ch->Q+ch->u_rows-ch->u_rank,
			  ch->u_rows, work_u+v_pos, 1, scalar_t(1.), tmp, 1, depth);
	  std::copy(tmp, tmp+ch->u_rows, ch->work_u);
	  delete[] tmp;
	} else std::copy(work_u+v_pos, work_u+v_pos+ch->u_rank, ch->work_u+ch->u_rows-ch->u_rank);
	v_pos += ch->u_rank;
      }

      if (tasked) {
	for (auto ch : child)
#pragma omp task firstprivate(ch) default(shared) final(depth >= params::task_recursion_cutoff_level-1) mergeable
	  ch->ULV_backward_solve(f, depth+1, false);
#pragma omp taskwait
      } else {
	for (auto ch : child) ch->ULV_backward_solve(f, depth, false);
      }
    } else std::copy(work_u, work_u+u_rows, f+subset_begin);
  }

} // end namespace strumpack

#endif //HSS_MATRIX_ULV_HPP
