/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef HSS_MATRIX_MATVEC_HPP
#define HSS_MATRIX_MATVEC_HPP

namespace strumpack {

  // Martinsson algorithm 1:  y = Ahss * x + beta * y
  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::matvec
  (scalar_t* x, integer_t ldx, scalar_t beta, scalar_t* y, integer_t ldy, integer_t nrhs, int depth) {
    bool tasked = depth<params::task_recursion_cutoff_level;
    if (child.empty()) {
      if (tasked) gemm_omp_task('N', 'N', u_rows, nrhs, u_rows, scalar_t(1.), D, u_rows,   // y = D x + beta y
				&x[subset_begin], ldx, beta, &y[subset_begin], ldy, depth);
      else blas::gemm('N', 'N', u_rows, nrhs, u_rows, scalar_t(1.), D, u_rows,   // y = D x + beta y
		      &x[subset_begin], ldx, beta, &y[subset_begin], ldy);
      return;
    }
    if (tasked) {
      for (auto ch : child) {
#pragma omp task firstprivate(ch) default(shared) final(depth >= params::task_recursion_cutoff_level-1) mergeable
	ch->matvec_forward_loop(x, ldx, nrhs, depth+1);
      }
#pragma omp taskwait
    } else for (auto ch : child) ch->matvec_forward_loop(x, ldx, nrhs, depth);
    matvec_backward_loop(x, ldx, beta, y, ldy, nrhs, true, depth);
  }

  // Based on Martinsson algorithm 1:  y = transpose(Ahss) * x + beta * y
  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::matvec_transpose
  (scalar_t* x, integer_t ldx, scalar_t beta, scalar_t* y, integer_t ldy, integer_t nrhs, int depth) {
    bool tasked = depth<params::task_recursion_cutoff_level;
    if (child.empty()) {
      if (tasked) gemm_omp_task('C', 'N', u_rows, nrhs, u_rows, scalar_t(1.), D, u_rows,   // y = trans(D) x + y
				&x[subset_begin], ldx, beta, &y[subset_begin], ldy, depth);
      else blas::gemm('C', 'N', u_rows, nrhs, u_rows, scalar_t(1.), D, u_rows,   // y = trans(D) x + y
		      &x[subset_begin], ldx, beta, &y[subset_begin], ldy);
      return;
    }
    if (tasked) {
      for (auto ch : child) {
#pragma omp task firstprivate(ch) default(shared) final(depth >= params::task_recursion_cutoff_level-1) mergeable
	ch->matvec_transpose_forward_loop(x, ldx, nrhs, depth+1);
      }
#pragma omp taskwait
    } else for (auto ch : child) ch->matvec_transpose_forward_loop(x, ldx, nrhs, depth);
    matvec_transpose_backward_loop(x, ldx, beta, y, ldy, nrhs, true, depth);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::matvec_forward_loop
  (scalar_t* x, integer_t ldx, integer_t nrhs, int depth) {
    reallocate_workmemory(u_rank*nrhs, v_rank*nrhs);
    bool tasked = depth<params::task_recursion_cutoff_level;

    if (child.empty()) { // leaf
      apply_V(true, nrhs, &x[subset_begin], ldx, scalar_t(0.), work_v, v_rank, depth);  // work_v = V^T x
    } else { // nonleaf
      if (tasked) {
	for (auto ch : child)
#pragma omp task firstprivate(ch) default(shared) final(depth >= params::task_recursion_cutoff_level-1) mergeable
	  ch->matvec_forward_loop(x, ldx, nrhs, depth+1);
#pragma omp taskwait
      } else {
	for (auto ch : child) ch->matvec_forward_loop(x, ldx, nrhs, depth);
      }

      if (v_rank) {
	scalar_t* tmp = new scalar_t[v_rows*nrhs];
	integer_t pos = 0;
	for (auto ch : child) { // tmp = [ work_v_nu1; work_v_nu2; .. ]
	  blas::lacpy('A', ch->v_rank, nrhs, ch->work_v, ch->v_rank, &tmp[pos], v_rows);
	  pos += ch->v_rank;
	}
	apply_V(true, nrhs, tmp, v_rows, scalar_t(0.), work_v, v_rank, depth);
	delete[] tmp;
      }
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::matvec_backward_loop
  (scalar_t* x, integer_t ldx, scalar_t beta, scalar_t* y, integer_t ldy, integer_t nrhs, bool isroot, int depth) {
    bool tasked = depth<params::task_recursion_cutoff_level;
    if (child.empty()) { // leaf, cannot be root because that special case is already handled
      if (u_rank) {
	apply_U(false, nrhs, work_u, u_rank, beta, &y[subset_begin], ldy, depth);    // y = U work_u + beta y
	if (tasked) gemm_omp_task('N', 'N', u_rows, nrhs, u_rows, scalar_t(1.), D, u_rows,   // y = D x + y
				  &x[subset_begin], ldx, scalar_t(1.), &y[subset_begin], ldy, depth);
	else blas::gemm('N', 'N', u_rows, nrhs, u_rows, scalar_t(1.), D, u_rows,   // y = D x + y
			&x[subset_begin], ldx, scalar_t(1.), &y[subset_begin], ldy);
      } else {
	if (tasked) gemm_omp_task('N', 'N', u_rows, nrhs, u_rows, scalar_t(1.), D, u_rows,   // y = D x + beta y
				  &x[subset_begin], ldx, beta, &y[subset_begin], ldy, depth);
	else blas::gemm('N', 'N', u_rows, nrhs, u_rows, scalar_t(1.), D, u_rows,   // y = D x + beta y
			&x[subset_begin], ldx, beta, &y[subset_begin], ldy);
      }
    } else {
      if (!isroot && u_rank) {
	scalar_t* tmp = new scalar_t[u_rows*nrhs];
	integer_t pos = 0;
	apply_U(false, nrhs, work_u, u_rank, scalar_t(0.), tmp, u_rows, depth);
	for (auto ch : child) { // tmp = [ work_v_nu1; work_v_nu2; .. ]
	  blas::lacpy('A', ch->u_rank, nrhs, &tmp[pos], u_rows, ch->work_u, ch->u_rank);
	  pos += ch->u_rank;
	}
	delete[] tmp;
      } else
	for (auto ch : child)
	  mat_zero(ch->u_rank, nrhs, ch->work_u, ch->u_rank);

      integer_t Bpos = 0;
      for (auto ci : child)
	for (auto cj : child)
	  if (ci != cj && cj->u_rank>0 && ci->v_rank>0) {
	    if (tasked) gemm_omp_task('N', 'N', cj->u_rank, nrhs, ci->v_rank, scalar_t(1.0), &B[Bpos], cj->u_rank,
				      ci->work_v, ci->v_rank, scalar_t(1.), cj->work_u, cj->u_rank, depth);
	    else blas::gemm('N', 'N', cj->u_rank, nrhs, ci->v_rank, scalar_t(1.0), &B[Bpos], cj->u_rank,
			    ci->work_v, ci->v_rank, scalar_t(1.), cj->work_u, cj->u_rank);
	    Bpos += cj->u_rank * ci->v_rank;
	  }

      if (tasked) {
	for (auto ch : child)
#pragma omp task firstprivate(ch) default(shared) final(depth >= params::task_recursion_cutoff_level-1) mergeable
	  ch->matvec_backward_loop(x, ldx, beta, y, ldy, nrhs, false, depth+1);
#pragma omp taskwait
      } else {
	for (auto ch : child) ch->matvec_backward_loop(x, ldx, beta, y, ldy, nrhs, false, depth);
      }
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::matvec_transpose_forward_loop
  (scalar_t* x, integer_t ldx, integer_t nrhs, int depth) {
    reallocate_workmemory(u_rank*nrhs, v_rank*nrhs);
    bool tasked = depth<params::task_recursion_cutoff_level;

    if (child.empty()) { // leaf
      apply_U(true, nrhs, &x[subset_begin], ldx, scalar_t(0.), work_u, u_rank, depth);  // work_u = U^T x
    } else { // nonleaf
      if (tasked) {
	for (auto ch : child)
#pragma omp task firstprivate(ch) default(shared) final(depth >= params::task_recursion_cutoff_level-1) mergeable
	  ch->matvec_transpose_forward_loop(x, ldx, nrhs, depth+1);
#pragma omp taskwait
      } else {
	for (auto ch : child) ch->matvec_transpose_forward_loop(x, ldx, nrhs, depth);
      }
      if (u_rank) {
	scalar_t* tmp = new scalar_t[u_rows*nrhs];
	integer_t pos = 0;
	for (auto ch : child) { // tmp = [ work_u_nu1; work_u_nu2; .. ]
	  blas::lacpy('A', ch->u_rank, nrhs, ch->work_u, ch->u_rank, &tmp[pos], u_rows);
	  pos += ch->u_rank;
	}
	apply_U(true, nrhs, tmp, u_rows, scalar_t(0.), work_u, u_rank, depth);
	delete[] tmp;
      }
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::matvec_transpose_backward_loop
  (scalar_t* x, integer_t ldx, scalar_t beta, scalar_t* y, integer_t ldy, integer_t nrhs, bool isroot, int depth) {
    bool tasked = depth<params::task_recursion_cutoff_level;
    if (child.empty()) { // leaf, cannot be root because that special case is already handled
      if (v_rank) {
	apply_V(false, nrhs, work_v, v_rank, beta, &y[subset_begin], ldy, depth);     // y = V work_v + beta y
	if (tasked) gemm_omp_task('C', 'N', v_rows, nrhs, v_rows, scalar_t(1.0), D, v_rows,   // y = trans(D) x + y
				  &x[subset_begin], ldx, scalar_t(1.), &y[subset_begin], ldy, depth);
	else blas::gemm('C', 'N', v_rows, nrhs, v_rows, scalar_t(1.0), D, v_rows,   // y = trans(D) x + y
			&x[subset_begin], ldx, scalar_t(1.), &y[subset_begin], ldy);
      } else {
	if (tasked) gemm_omp_task('C', 'N', v_rows, nrhs, v_rows, scalar_t(1.0), D, v_rows,   // y = trans(D) x + beta y
				  &x[subset_begin], ldx, beta, &y[subset_begin], ldy, depth);
	else blas::gemm('C', 'N', v_rows, nrhs, v_rows, scalar_t(1.0), D, v_rows,   // y = trans(D) x + beta y
			&x[subset_begin], ldx, beta, &y[subset_begin], ldy);
      }
    } else {
      if (!isroot && v_rank) {
	scalar_t* tmp = new scalar_t[v_rows*nrhs];
	integer_t pos = 0;
	apply_V(false, nrhs, work_v, v_rank, scalar_t(0.), tmp, v_rows, depth);
	for (auto ch : child) { // tmp = [ work_v_nu1; work_v_nu2; .. ]
	  blas::lacpy('A', ch->v_rank, nrhs, &tmp[pos], v_rows, ch->work_v, ch->v_rank);
	  pos += ch->v_rank;
	}
	delete[] tmp;
      } else
	for (auto ch : child)
	  mat_zero(ch->v_rank, nrhs, ch->work_v, ch->v_rank);

      integer_t Bpos = 0;
      for (auto cj : child)
	for (auto ci : child)
	  if (ci != cj && cj->v_rank>0 && ci->u_rank>0) {
	    if (tasked) gemm_omp_task('C', 'N', cj->v_rank, nrhs, ci->u_rank, scalar_t(1.0), &B[Bpos], ci->u_rank,
				      ci->work_u, ci->u_rank, scalar_t(1.), cj->work_v, cj->v_rank, depth);
	    else blas::gemm('C', 'N', cj->v_rank, nrhs, ci->u_rank, scalar_t(1.0), &B[Bpos], ci->u_rank,
			    ci->work_u, ci->u_rank, scalar_t(1.), cj->work_v, cj->v_rank);
	    Bpos += cj->v_rank * ci->u_rank;
	  }

      if (tasked) {
	for (auto ch : child)
#pragma omp task firstprivate(ch) default(shared) final(depth >= params::task_recursion_cutoff_level-1) mergeable
	  ch->matvec_transpose_backward_loop(x, ldx, beta, y, ldy, nrhs, false, depth+1);
#pragma omp taskwait
      } else {
	for (auto ch : child) ch->matvec_transpose_backward_loop(x, ldx, beta, y, ldy, nrhs, false, depth);
      }
    }
  }

} // end namespace strumpack

#endif //HSS_MATRIX_MATVEC_HPP
