/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef MATRIX_REORDERING_HPP
#define MATRIX_REORDERING_HPP

#include <vector>
#include <string>
#include <algorithm>
#include <unordered_map>
#include <scotch.h>
#include <metis.h>
#include "strumpack_parameters.hpp"
#include "CompressedSparseMatrix.hpp"
#include "MPI_wrapper.hpp"

namespace strumpack {

  /** a node in this tree should always have 0 or 2 children */
  class HSSPartitionTree {
  public:
    int size;
    std::vector<HSSPartitionTree> c;
    HSSPartitionTree() : size(0) {}
    HSSPartitionTree(int n) : size(n) {}
    HSSPartitionTree(std::vector<int>& buf) {
      int n = buf.size() / 3;
      int pid = n-1;
      de_serialize_rec(buf.data(), buf.data()+n, buf.data()+2*n, pid);
    }
    HSSPartitionTree(int buf_size, int* buf) {
      int n = buf_size / 3;
      int pid = n-1;
      de_serialize_rec(buf, buf+n, buf+2*n, pid);
    }
    void refine(int leaf_size) {
      assert(c.empty());
      if (size > 2*leaf_size) {
	c.resize(2);
	c[0].size = size/2;
	c[0].refine(leaf_size);
	c[1].size = size - size/2;
	c[1].refine(leaf_size);
      }
    }
    void print() {
      for (auto& ch : c) ch.print();
      std::cout << size << " ";
    }
    int nodes() {
      int nr_nodes = 1;
      for (auto& ch : c) nr_nodes += ch.nodes();
      return nr_nodes;
    }
    std::vector<int> serialize() {
      int n = nodes(), pid = 0;
      std::vector<int> buf(3*n);
      serialize_rec(buf.data(), buf.data()+n, buf.data()+2*n, pid);
      return buf;
    }
  private:
    void serialize_rec(int* sizes, int* lchild, int* rchild, int& pid) {
      if (!c.empty()) {
	c[0].serialize_rec(sizes, lchild, rchild, pid);
	auto lroot = pid;
	c[1].serialize_rec(sizes, lchild, rchild, pid);
	lchild[pid] = lroot-1;
	rchild[pid] = pid-1;
      } else lchild[pid] = rchild[pid] = -1;
      sizes[pid++] = size;
    }
    void de_serialize_rec(int* sizes, int* lchild, int* rchild, int& pid) {
      size = sizes[pid--];
      if (rchild[pid+1] != -1) {
	c.resize(2);
	c[1].de_serialize_rec(sizes, lchild, rchild, pid);
	c[0].de_serialize_rec(sizes, lchild, rchild, pid);
      }
    }
  };

  template<typename integer_t> std::vector<int>
  serialize(std::unordered_map<integer_t, HSSPartitionTree>& hss_tree_map) {
    std::vector<int> buf;
    buf.push_back(hss_tree_map.size());
    for (auto& ht : hss_tree_map) {
      buf.push_back(ht.first);
      auto ht_buf = ht.second.serialize();
      buf.push_back(ht_buf.size());
      buf.insert(buf.end(), ht_buf.begin(), ht_buf.end());
    }
    return buf;
  }

  template<typename scalar_t,typename real_t,typename integer_t> class MatrixReordering {
  public:
    integer_t nbsep;
    std::vector<integer_t> order;
    std::vector<integer_t> iorder;
    std::vector<integer_t> sep_ptr;
    std::vector<integer_t> parent;
    std::vector<integer_t> left_child;
    std::vector<integer_t> right_child;
    std::vector<integer_t> level;

    std::vector<integer_t> sep_order;
    std::vector<integer_t> sep_iorder;
    std::unordered_map<integer_t, HSSPartitionTree> sep_HSS_tree;

    MatrixReordering() : nbsep(0) {}
    virtual ~MatrixReordering() {}

    virtual int nested_dissection(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A,
				  bool diag, int stratnum, int stratpar, bool verbose) {
      return nested_dissection(A, diag, stratnum, stratpar, verbose, 0, 0, 0);
    }
    virtual int nested_dissection(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A,
				  bool diag, int stratnum, int stratpar, bool verbose,
				  int nx, int ny, int nz) = 0;

    virtual void separator_reordering(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool verbose);
    virtual std::string get_name() = 0;
    virtual void clear_tree_data();
    void broadcast_nd(MPI_Comm comm);

  protected:
    virtual void split_separator(HSSPartitionTree& hss_tree, integer_t& nr_parts, integer_t sep,
				 CompressedSparseMatrix<scalar_t,real_t,integer_t>* A,
				 integer_t part, integer_t count);
    void extract_separator(integer_t part, integer_t sep_beg, integer_t sep_end,
			   CompressedSparseMatrix<scalar_t,real_t,integer_t>* A,
			   std::vector<idx_t>& xadj, std::vector<idx_t>& adjncy);
    void separator_reordering_recursive(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A,
					bool hss_parent, integer_t sep);
    virtual void nested_dissection_print(integer_t n, integer_t nnz, int stratnum, int stratpar, bool verbose);
    void separator_reordering_print(integer_t max_nr_neighbours, integer_t max_dim_sep);
  };

  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReordering<scalar_t,real_t,integer_t>::clear_tree_data() {
    nbsep = 0;
    std::vector<integer_t>().swap(sep_ptr);
    std::vector<integer_t>().swap(parent);
    std::vector<integer_t>().swap(left_child);
    std::vector<integer_t>().swap(right_child);
    std::vector<integer_t>().swap(level);
    sep_HSS_tree.clear();
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReordering<scalar_t,real_t,integer_t>::broadcast_nd(MPI_Comm comm) {
    integer_t n = order.size();
    MPI_Bcast(&n,     1, mpi_type<integer_t>(), 0, comm);
    MPI_Bcast(&nbsep, 1, mpi_type<integer_t>(), 0, comm);
    size_t N = 2+2*n+5*nbsep+1;
    auto buf = new integer_t[N];
    auto pbuf = buf;
    if (!mpi_rank(comm)) {
      buf[0] = nbsep; pbuf += 1;
      buf[1] = n;     pbuf += 1;
      std::copy(order.begin(),       order.end(),       pbuf); pbuf += order.size();
      std::copy(iorder.begin(),      iorder.end(),      pbuf); pbuf += iorder.size();
      std::copy(sep_ptr.begin(),     sep_ptr.end(),     pbuf); pbuf += sep_ptr.size();
      std::copy(parent.begin(),      parent.end(),      pbuf); pbuf += parent.size();
      std::copy(left_child.begin(),  left_child.end(),  pbuf); pbuf += left_child.size();
      std::copy(right_child.begin(), right_child.end(), pbuf); pbuf += right_child.size();
      std::copy(level.begin(),       level.end(),       pbuf);
    }
    MPI_Bcast(buf, N, mpi_type<integer_t>(), 0, comm);
    if (mpi_rank(comm)) {
      nbsep = buf[0]; pbuf += 1;
      n     = buf[1]; pbuf += 1;
      order.resize(n);
      iorder.resize(n);
      sep_ptr.resize(nbsep+1);
      parent.resize(nbsep);
      left_child.resize(nbsep);
      right_child.resize(nbsep);
      level.resize(nbsep);
      std::copy(pbuf, pbuf+order.size(),       order.begin());       pbuf += order.size();
      std::copy(pbuf, pbuf+iorder.size(),      iorder.begin());      pbuf += iorder.size();
      std::copy(pbuf, pbuf+sep_ptr.size(),     sep_ptr.begin());     pbuf += sep_ptr.size();
      std::copy(pbuf, pbuf+parent.size(),      parent.begin());      pbuf += parent.size();
      std::copy(pbuf, pbuf+left_child.size(),  left_child.begin());  pbuf += left_child.size();
      std::copy(pbuf, pbuf+right_child.size(), right_child.begin()); pbuf += right_child.size();
      std::copy(pbuf, pbuf+level.size(),       level.begin());
    }
    delete[] buf;
  }

  // reorder the vertices in the separator to get a better rank structure
  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReordering<scalar_t,real_t,integer_t>::separator_reordering
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool verbose) {
    sep_order.resize(A->size());
    sep_iorder.resize(A->size());
#pragma omp parallel
#pragma omp single
    {
      separator_reordering_recursive(A, true, nbsep-1);
    }
    for (integer_t i=0; i<A->size(); i++) {
      sep_order[i] = -sep_order[i];
      sep_iorder[sep_order[i]] = i;
    }
    //if (verbose) separator_reordering_print(max_nr_neighbours, max_dim_sep);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReordering<scalar_t,real_t,integer_t>::separator_reordering_recursive
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool hss_parent, integer_t sep) {
    auto sep_begin = sep_ptr[sep];
    auto sep_end = sep_ptr[sep + 1];
    auto dim_sep = sep_end - sep_begin;
    bool is_hss = hss_parent && (dim_sep >= params::hss_minimum_front_size);
#pragma omp task if(is_hss)
    if (left_child[sep] != -1) separator_reordering_recursive(A, is_hss, left_child[sep]);
#pragma omp task if(is_hss)
    if (right_child[sep] != -1) separator_reordering_recursive(A, is_hss, right_child[sep]);
    if (is_hss) {
      HSSPartitionTree tree(dim_sep);
      if (dim_sep > 2 * params::hss_minimum_partition_size) {
#pragma omp task
	{
	  integer_t nr_parts = 0;
	  split_separator(tree, nr_parts, sep, A, 0, 1);
	  auto count = sep_begin;
	  for (integer_t part=0; part<nr_parts; part++)
	    for (integer_t i=sep_begin; i<sep_end; i++)
	      if (sep_order[i] == part) sep_order[i] = -count++;
	}
      } else for (integer_t i=sep_begin; i<sep_end; i++) sep_order[i] = -i;
#pragma omp critical
      {
	sep_HSS_tree[sep] = tree; // not thread safe!
      }
    } else for (integer_t i=sep_begin; i<sep_end; i++) sep_order[i] = -i;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReordering<scalar_t,real_t,integer_t>::split_separator
  (HSSPartitionTree& hss_tree, integer_t& nr_parts, integer_t sep,
   CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, integer_t part, integer_t count) {
    auto sep_begin = sep_ptr[sep];
    auto sep_end = sep_ptr[sep+1];
    std::vector<idx_t> xadj, adjncy;
    extract_separator(part, sep_begin, sep_end, A, xadj, adjncy);
    idx_t ncon = 1, edge_cut = 0, two = 2, nvtxs=xadj.size()-1;
    auto partitioning = new idx_t[nvtxs];
    int info = METIS_PartGraphRecursive(&nvtxs, &ncon, xadj.data(), adjncy.data(), NULL, NULL, NULL,
					&two, NULL, NULL, NULL, &edge_cut, partitioning);
    if (info != METIS_OK) {
      std::cerr << "METIS_PartGraphRecursive for separator reordering returned: " << info << std::endl;
      exit(1);
    }
    hss_tree.c.resize(2);
    for (integer_t i=sep_begin, j=0; i<sep_end; i++)
      if (sep_order[i] == part) {
	auto p = partitioning[j++];
	sep_order[i] = -count - p;
	hss_tree.c[p].size++;
      }
    delete[] partitioning;
    for (integer_t p=0; p<2; p++)
      if (hss_tree.c[p].size > 2 * params::hss_minimum_partition_size)
	split_separator(hss_tree.c[p], nr_parts, sep, A, -count-p, count+2);
      else std::replace(&sep_order[sep_begin], &sep_order[sep_end], -count-p, nr_parts++);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReordering<scalar_t,real_t,integer_t>::extract_separator
  (integer_t part, integer_t sep_begin, integer_t sep_end, CompressedSparseMatrix<scalar_t,real_t,integer_t>* A,
   std::vector<idx_t>& xadj, std::vector<idx_t>& adjncy) {
    assert(params::separator_ordering_level == 0 || params::separator_ordering_level == 1);
    auto dim_sep = sep_end - sep_begin;
    auto mark = new bool[dim_sep];
    auto ind_to_part = new integer_t[dim_sep];
    integer_t nvtxs = 0;
    for (integer_t r=0; r<dim_sep; r++)
      ind_to_part[r] = (sep_order[r+sep_begin] == part) ? nvtxs++ : -1;
    xadj.reserve(nvtxs+1);
    adjncy.reserve(5*nvtxs);
    for (integer_t i=sep_begin, e=0; i<sep_end; i++) {
      if (sep_order[i] == part) {
	xadj.push_back(e);
	std::fill(mark, mark+dim_sep, false);
	for (integer_t j=A->get_ptr()[i]; j<A->get_ptr()[i+1]; j++) {
	  auto c = A->get_ind()[j];
	  if (c == i) continue;
	  auto lc = c - sep_begin;
	  if (lc >= 0 && lc < dim_sep && sep_order[c]==part && !mark[lc]) {
	    mark[lc] = true;
	    adjncy.push_back(ind_to_part[lc]);
	    e++;
	  } else {
	    if (params::separator_ordering_level > 0) {
	      for (integer_t k=A->get_ptr()[c]; k<A->get_ptr()[c+1]; k++) {
	   	auto cc = A->get_ind()[k];
	   	auto lcc = cc - sep_begin;
	   	if (cc!=i && lcc >= 0 && lcc < dim_sep && sep_order[cc]==part && !mark[lcc]) {
	   	  mark[lcc] = true;
	   	  adjncy.push_back(ind_to_part[lcc]);
		  e++;
	   	}
	      }
	    }
	  }
	}
      }
    }
    xadj.push_back(adjncy.size());
    delete[] mark;
    delete[] ind_to_part;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReordering<scalar_t,real_t,integer_t>::nested_dissection_print
  (integer_t n, integer_t nnz, int stratnum, int stratpar, bool verbose) {
    if (verbose) {
      std::cout << "# initial matrix:" << std::endl;
      std::cout << "#   - number of unknowns = " << n << std::endl;
      std::cout << "#   - number of nonzeros = " << nnz << std::endl;
      std::cout << "# nested dissection reordering:" << std::endl;
      std::cout << "#   - " << get_name() << " reordering" << std::endl;
      std::cout << "#   - strategy number = " << stratnum << std::endl;
      std::cout << "#   - strategy parameter = " << stratpar << std::endl;
      std::cout << "#   - number of separators = " << nbsep << std::endl;
      std::cout << "#   - number of levels = " << level.back() << std::endl;
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReordering<scalar_t,real_t,integer_t>::separator_reordering_print
  (integer_t max_nr_neighbours, integer_t max_dim_sep) {
    std::cout << "# separator reordering:" << std::endl;
    std::cout << "#   - maximum connectivity = " << max_nr_neighbours << std::endl;
    std::cout << "#   - maximum separator dimension = " << max_dim_sep << std::endl;
  }

} // end namespace strumpack

#endif
