/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government igs granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef RANDOM_WRAPPER_HPP
#define RANDOM_WRAPPER_HPP

#include <memory>
#include <random>

namespace strumpack {

  template<typename real_t>
  class RandomGeneratorBase {
  public:
    virtual void seed(size_t s) = 0;
    virtual real_t get() = 0;
    virtual int flops_per_prng() = 0;
  };

  template<typename real_t, typename E, typename D>
  class RandomGenerator : public RandomGeneratorBase<real_t> {
  public:
    RandomGenerator() {}
    RandomGenerator(size_t s) { e.seed(s); }
    void seed(size_t s) { e.seed(s); d.reset(); }
    real_t get() { return d(e); }
    int flops_per_prng() {
      if (std::is_same<D,std::normal_distribution<real_t>>()) return 23;
      if (std::is_same<D,std::uniform_real_distribution<real_t>>()) return 7;
      return 0;
    }
  private:
    D d;
    E e;
  };

  template<typename real_t>
  std::unique_ptr<RandomGeneratorBase<real_t>> make_random_generator(size_t seed) {
    if (params::random_engine == params::LINEAR) {
      if (params::random_distribution == params::NORMAL)
	return std::unique_ptr<RandomGeneratorBase<real_t>>
	  (new RandomGenerator<real_t,std::minstd_rand,std::normal_distribution<real_t>>(seed));
      else if (params::random_distribution == params::UNIFORM)
	return std::unique_ptr<RandomGeneratorBase<real_t>>
	  (new RandomGenerator<real_t,std::minstd_rand,std::uniform_real_distribution<real_t>>(seed));
    } else if (params::random_engine == params::MERSENNE) {
      if (params::random_distribution == params::NORMAL)
	return std::unique_ptr<RandomGeneratorBase<real_t>>
	  (new RandomGenerator<real_t,std::mt19937,std::normal_distribution<real_t>>(seed));
      else if (params::random_distribution == params::UNIFORM)
	return std::unique_ptr<RandomGeneratorBase<real_t>>
	  (new RandomGenerator<real_t,std::mt19937,std::uniform_real_distribution<real_t>>(seed));
    }
    return NULL;
  }

  template<typename real_t>
  std::unique_ptr<RandomGeneratorBase<real_t>> make_random_generator() {
    if (params::random_engine == params::LINEAR) {
      if (params::random_distribution == params::NORMAL)
  	return std::unique_ptr<RandomGeneratorBase<real_t>>
  	  (new RandomGenerator<real_t,std::minstd_rand,std::normal_distribution<real_t>>());
      else if (params::random_distribution == params::UNIFORM)
  	return std::unique_ptr<RandomGeneratorBase<real_t>>
  	  (new RandomGenerator<real_t,std::minstd_rand,std::uniform_real_distribution<real_t>>());
    } else if (params::random_engine == params::MERSENNE) {
      if (params::random_distribution == params::NORMAL)
  	return std::unique_ptr<RandomGeneratorBase<real_t>>
  	  (new RandomGenerator<real_t,std::mt19937,std::normal_distribution<real_t>>());
      else if (params::random_distribution == params::UNIFORM)
  	return std::unique_ptr<RandomGeneratorBase<real_t>>
  	  (new RandomGenerator<real_t,std::mt19937,std::uniform_real_distribution<real_t>>());
    }
    return NULL;
  }

}

#endif // RANDOM_WRAPPER
