/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 */
/*! \file StrumpackSparseSolverMPI.hpp
 * \brief Contains the definition of MPI distributed sparse solver class with REPLICATED input.
 */
#ifndef STRUMPACK_SPARSE_SOLVER_MPI_H
#define STRUMPACK_SPARSE_SOLVER_MPI_H

#include "StrumpackSparseSolver.hpp"
#include "EliminationTreeMPI.hpp"
#include "mpi.h"

namespace strumpack {

  /*! \brief This is the interface for the distributed memory solver
   *         with REPLICATED input.
   *
   * The factorization and solve are distributed using MPI. The sparse
   * factors (fill-in) are distributed over the MPI processes, but
   * each MPI process needs a copy of the entire input sparse matrix,
   * the entire right-hand side and the entire solution vector. The
   * matrix reordering and the symbolic factorization are not
   * distributed (but they are threaded).
   *
   * The StrumpackSparseSolverMPI class depends on 3 template
   * paramaters <scalar_t,real_t,integer_t>. Supported combinations
   * for scalar_t,real_t are: <float,float>, <double,double>,
   * <std::complex<float>,float> and
   * <std::complex<double>,double>. The integer_t type defaults to a
   * regular int. If regular int causes 32 bit integer overflows, you
   * should switch to integer_t=int64_t instead.
   */
  template<typename scalar_t,typename real_t,typename integer_t>
  class StrumpackSparseSolverMPI : public StrumpackSparseSolver<scalar_t,real_t,integer_t> {
  public:
    /*! \brief Constructor taking an MPI communicator and command line
     *         arguments.
     *
     * \param mpi_comm MPI communicator. Can be MPI_COMM_WORLD or a subcommunicator.
     * \param argc  The number of arguments, i.e, number of elements in the argv array.
     * \param argv  Command line arguments. Add -h or --help to have a description printed.
     * \param verb  Flag to suppres/enable output. Only the root of mpi_comm will print output to stdout.
     */
    StrumpackSparseSolverMPI(MPI_Comm mpi_comm, int argc, char* argv[], bool verb=true);
    /*! \brief Destructor. */
    virtual ~StrumpackSparseSolverMPI() { mpi_free_comm(&comm); }
    virtual params::returnCode reorder(int nx=1, int ny=1, int nz=1);

  protected:
    EliminationTree<scalar_t,real_t,integer_t>* get_elimination_tree();
    void perf_counters_stop(std::string s);
    virtual void synchronize() { MPI_Barrier(comm); }

    MPI_Comm comm;
  };

  template<typename scalar_t,typename real_t,typename integer_t>
  StrumpackSparseSolverMPI<scalar_t,real_t,integer_t>::StrumpackSparseSolverMPI
  (MPI_Comm mpi_comm, int argc, char* argv[], bool verb) :
    StrumpackSparseSolver<scalar_t,real_t,integer_t>(argc, argv, verb, mpi_rank(mpi_comm) == 0) {
    MPI_Comm_dup(mpi_comm, &comm);
    if (this->verbose && this->is_root)
      std::cout << "# using " << mpi_nprocs(comm) << " MPI processes" << std::endl;
  }

  template<typename scalar_t,typename real_t,typename integer_t>
  EliminationTree<scalar_t,real_t,integer_t>*
  StrumpackSparseSolverMPI<scalar_t,real_t,integer_t>::get_elimination_tree() {
    return new EliminationTreeMPI<scalar_t,real_t,integer_t>(this->mat, this->nd, this->hss_strategy.use_hss, comm);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolverMPI<scalar_t,real_t,integer_t>::perf_counters_stop(std::string s) {
    if (this->verbose) {
#if defined(HAVE_PAPI)
      float rtime1=0., ptime1=0., mflops=0.;
      long_long flpops1=0;
#pragma omp parallel reduction(+:flpops1) reduction(max:rtime1) reduction(max:ptime1)
      PAPI_flops(&rtime1, &ptime1, &flpops1, &mflops);
      float papi_total_flops = flpops1 - this->flpops0;
      MPI_Allreduce(MPI_IN_PLACE, &papi_total_flops, 1, MPI_FLOAT, MPI_SUM, comm);
      if (this->is_root)
	std::cout << "# " << s << " PAPI stats:" << std::endl
		  << "#   - total flops = " << papi_total_flops << std::endl
		  << "#   - flop rate = " <<  papi_total_flops/(rtime1-this->rtime0)/1e9 << " GFlops/s" << std::endl
		  << "#   - real time = " << rtime1-this->rtime0 << " sec" << std::endl
		  << "#   - processor time = " << ptime1-this->ptime0 << " sec" << std::endl;
#endif
#if defined(COUNT_FLOPS)
      long long int flopsbytes[2] = { params::flops - this->f0, params::bytes - this->b0 };
      MPI_Allreduce(MPI_IN_PLACE, flopsbytes, 2, MPI_LONG_LONG_INT, MPI_SUM, comm);
      this->flops_total = flopsbytes[0];
      this->bytes_total = flopsbytes[1];
#endif
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> params::returnCode
  StrumpackSparseSolverMPI<scalar_t,real_t,integer_t>::reorder(int nx, int ny, int nz) {
    if (this->mat == NULL) return params::MATRIX_NOT_SET;
    TaskTimer t1("permute-scale");
    if (params::mc64job != 0) {
      if (this->verbose && this->is_root) {
	std::cout << "# mc64ad job " << params::mc64job << ": ";
	switch (params::mc64job) {
	case 1: std::cout << "maximum cardinality ! Doesn't work" << std::endl;
	  return params::REORDERING_ERROR; break;
	case 2: std::cout << "maximum smallest diagonal value, version 1" << std::endl; break;
	case 3: std::cout << "maximum smallest diagonal value, version 2" << std::endl; break;
	case 4: std::cout << "maximum sum of diagonal values" << std::endl; break;
	case 5: std::cout << "maximum matching + row and column scaling" << std::endl; break;
	}
      }
      t1.time([&](){ this->mat->permute_and_scale(params::mc64job, this->mc64_cperm, this->mc64_Dr, this->mc64_Dc); });
    }
    auto old_nnz = this->mat->nnz();
    TaskTimer t2("sparsity-symmetrization", [&](){ this->mat->symmetrize_sparsity(); });
    if (this->mat->nnz() != old_nnz && this->verbose && this->is_root)
      std::cout << "# Matrix padded with zeros to get symmetric pattern." << std::endl
		<< "# Number of nonzeros increased from "
		<< old_nnz << " to " << this->mat->nnz() << "." << std::endl;

    TaskTimer t3("nested-dissection");
    this->perf_counters_start();
    t3.start();
    if (this->nd) delete this->nd;
    this->nd = this->get_matrix_reordering();
    if (!mpi_rank(comm)) {
      int ierr = this->nd->nested_dissection
	(this->mat, true, this->scotch_strategy, this->nested_dissection_parameter,
	 this->verbose && this->is_root, nx, ny, nz);
      if (ierr) {
	std::cerr << "ERROR: nested dissection went wrong, ierr=" << ierr << std::endl;
	return params::REORDERING_ERROR;
      }
    }
    this->nd->broadcast_nd(comm);
    this->mat->permute(this->nd->iorder, this->nd->order);
    t3.stop();
    if (this->verbose && this->is_root) {
      std::cout << "#   - nd time = " << t3.elapsed() << std::endl;
      if (params::mc64job != 0) std::cout << "#   - mc64 time = " << t1.elapsed() << std::endl;
      std::cout << "#   - symmetrization time = " << t2.elapsed() << std::endl;
    }
    perf_counters_stop("nested dissection");

    if (this->hss_strategy.use_hss) {
      this->perf_counters_start();
      TaskTimer t4("separator-reordering", [&](){
	  this->nd->separator_reordering(this->mat, this->verbose && this->is_root);
	  // TODO also broadcast this?? is computed with scotch
	  this->mat->permute(this->nd->sep_iorder, this->nd->sep_order);
	});
      if (this->verbose && this->is_root) std::cout << "#   - sep-reorder time = " << t4.elapsed() << std::endl;
      perf_counters_stop("separator reordering");
    }

    this->perf_counters_start();
    TaskTimer t0("symbolic-factorization", [&](){
	if (this->et) delete this->et;
	this->et = get_elimination_tree();
      });
    if (this->verbose) {
      auto nr_dense = this->et->get_nr_dense_fronts();
      auto nr_HSS = this->et->get_nr_HSS_fronts();
      if (this->is_root)
	std::cout << "# symbolic factorization:" << std::endl
		  << "#   - HSS compression = " << std::boolalpha << this->hss_strategy.use_hss << std::endl
		  << "#   - nr of dense Frontal matrices = " << nr_dense << std::endl
		  << "#   - nr of HSS Frontal matrices = " << nr_HSS << std::endl
		  << "#   - symb-factor time = " << t0.elapsed() << std::endl;
    }
    perf_counters_stop("symbolic factorization");
    this->reordered = true;
    return params::SUCCESS;
  }

} // end namespace strumpack

#endif
